# Copyright (C) 2006 Novell Inc.  All rights reserved.
# This program is free software; it may be used, copied, modified
# and distributed under the terms of the GNU General Public Licence,
# either version 2, or (at your option) any later version.

import fnmatch
import getpass
import glob
import os
import re
import shutil
import subprocess
import sys
from tempfile import NamedTemporaryFile, mkdtemp
from typing import List
from typing import Optional
from urllib.parse import urlsplit
from urllib.request import URLError, HTTPError
from xml.etree import ElementTree as ET

from . import conf
from . import connection
from . import core
from . import oscerr
from .core import get_buildinfo, meta_exists, get_buildconfig, dgst
from .core import get_binarylist, get_binary_file, run_external, return_external, raw_input
from .fetch import Fetcher, OscFileGrabber, verify_pacs
from .meter import create_text_meter
from .util import cpio
from .util import archquery, debquery, packagequery, rpmquery
from .util import repodata
from .util.helper import decode_it
from .util.xml import xml_parse


change_personality = {
    'i686': 'linux32',
    'i586': 'linux32',
    'i386': 'linux32',
    'ppc': 'powerpc32',
    's390': 's390',
    'sparc': 'linux32',
    'sparcv8': 'linux32',
}

can_also_build = {
    'aarch64': ['aarch64'],  # only needed due to used heuristics in build parameter evaluation
    'armv6l': ['armv4l', 'armv5l', 'armv6l', 'armv5el', 'armv6el'],
    'armv7l': ['armv4l', 'armv5l', 'armv6l', 'armv7l', 'armv5el', 'armv6el', 'armv7el'],
    'armv5el': ['armv4l', 'armv5l', 'armv5el'],  # not existing arch, just for compatibility
    'armv6el': ['armv4l', 'armv5l', 'armv6l', 'armv5el', 'armv6el'],  # not existing arch, just for compatibility
    'armv6hl': ['armv4l', 'armv5l', 'armv6l', 'armv5el', 'armv6el'],
    'armv7el': ['armv4l', 'armv5l', 'armv6l', 'armv7l', 'armv5el', 'armv6el', 'armv7el'],  # not existing arch, just for compatibility
    'armv7hl': ['armv7hl'],  # not existing arch, just for compatibility
    'armv8el': ['armv4l', 'armv5el', 'armv6el', 'armv7el', 'armv8el'],  # not existing arch, just for compatibility
    'armv8l': ['armv4l', 'armv5el', 'armv6el', 'armv7el', 'armv8el'],  # not existing arch, just for compatibility
    'armv5tel': ['armv4l', 'armv5el', 'armv5tel'],
    's390x': ['s390'],
    'ppc64': ['ppc', 'ppc64', 'ppc64p7', 'ppc64le'],
    'ppc64le': ['ppc64le', 'ppc64'],
    'i586': ['i386'],
    'i686': ['i586', 'i386'],
    'x86_64': ['i686', 'i586', 'i386'],
    'sparc64': ['sparc64v', 'sparcv9v', 'sparcv9', 'sparcv8', 'sparc'],
    'parisc': ['hppa'],
}

# real arch of this machine
hostarch = os.uname()[4]
if hostarch == 'i686':  # FIXME
    hostarch = 'i586'

if hostarch == 'parisc':
    hostarch = 'hppa'


class Buildinfo:
    """represent the contents of a buildinfo file"""

    def __init__(self, filename, apiurl, buildtype='spec', localpkgs=None, binarytype='rpm'):
        localpkgs = localpkgs or []
        try:
            tree = xml_parse(filename)
        except ET.ParseError:
            print('could not parse the buildinfo:', file=sys.stderr)
            print(open(filename).read(), file=sys.stderr)
            sys.exit(1)

        root = tree.getroot()

        self.apiurl = apiurl

        if root.find('error') is not None:
            sys.stderr.write('buildinfo is broken... it says:\n')
            error = root.findtext("error")
            if error.startswith('unresolvable: '):
                sys.stderr.write('unresolvable: ')
                sys.stderr.write('\n     '.join(error[14:].split(',')))
            else:
                sys.stderr.write(error)
            sys.stderr.write('\n')
            sys.exit(1)

        if not (apiurl.startswith('https://') or apiurl.startswith('http://')):
            raise URLError('invalid protocol for the apiurl: \'%s\'' % apiurl)

        self.buildtype = buildtype
        self.binarytype = binarytype
        self.apiurl = apiurl

        # are we building .rpm or .deb?
        # XXX: shouldn't we deliver the type via the buildinfo?
        self.pacsuffix = 'rpm'
        if self.buildtype in ('dsc', 'collax', 'deb'):
            self.pacsuffix = 'deb'
        if self.buildtype == 'arch':
            self.pacsuffix = 'arch'
        if self.buildtype == 'livebuild':
            self.pacsuffix = 'deb'
        if self.buildtype == 'snapcraft':
            # atm ubuntu is used as base, but we need to be more clever when
            # snapcraft also supports rpm
            self.pacsuffix = 'deb'

        # The architectures become a bit mad ...
        # buildarch: The architecture of the build result      (host arch in GNU definition)
        # hostarch:  The architecture of the build environment (build arch in GNU defintion)
        # crossarch: Same as hostarch, but indicating that a sysroot with an incompatible architecture exists
        self.buildarch = root.findtext("arch")
        self.crossarch = root.findtext("crossarch")
        self.hostarch = root.findtext("hostarch")
        self.release = root.findtext("release")
        if conf.config['api_host_options'][apiurl]['downloadurl']:
            # Formerly, this was set to False, but we have to set it to True, because a large
            # number of repos in OBS are misconfigured and don't actually have repos setup - they
            # are API only.
            self.enable_cpio = True
            self.downloadurl = conf.config['api_host_options'][apiurl]['downloadurl'] + "/repositories"
            if conf.config['http_debug']:
                print("⚠️   setting dl_url to %s" % conf.config['api_host_options'][apiurl]['downloadurl'])
        else:
            self.enable_cpio = True
            self.downloadurl = root.get('downloadurl')

        self.debuginfo = 0
        if root.find('debuginfo') is not None:
            try:
                self.debuginfo = int(root.findtext("debuginfo"))
            except ValueError:
                pass

        self.deps = []
        self.projects = {}
        self.keys = []
        self.prjkeys = []
        self.pathes = []
        self.urls = {}
        self.modules = []
        for node in root.findall('module'):
            self.modules.append(node.text)
        for node in root.findall('bdep'):
            if node.find('sysroot'):
                p = Pac(node, self.buildarch, self.pacsuffix,
                        apiurl, localpkgs)
            else:
                pac_arch = self.crossarch
                if pac_arch is None:
                    pac_arch = self.buildarch
                p = Pac(node, pac_arch, self.pacsuffix,
                        apiurl, localpkgs)
            if p.project:
                self.projects[p.project] = 1
            self.deps.append(p)
        for node in root.findall('path'):
            # old simple list for compatibility
            # XXX: really old? This is currently used for kiwi builds
            self.pathes.append(node.get('project') + "/" + node.get('repository'))
            # a hash providing the matching URL for specific repos for newer OBS instances
            if node.get('url'):
                baseurl = node.get('url').replace('%', '%%')
                if conf.config['api_host_options'][apiurl]['downloadurl']:
                    # Add the path element to the download url override.
                    baseurl = conf.config['api_host_options'][apiurl]['downloadurl'] + urlsplit(node.get('url'))[2]
                self.urls[node.get('project') + "/" + node.get('repository')] = baseurl + '/%(arch)s/%(filename)s'

        self.vminstall_list = [dep.name for dep in self.deps if dep.vminstall]
        self.preinstall_list = [dep.name for dep in self.deps if dep.preinstall]
        self.runscripts_list = [dep.name for dep in self.deps if dep.runscripts]
        self.noinstall_list = [dep.name for dep in self.deps if dep.noinstall]
        self.installonly_list = [dep.name for dep in self.deps if dep.installonly]

        if root.find('preinstallimage') is not None:
            self.preinstallimage = root.find('preinstallimage')
        else:
            self.preinstallimage = None

        self.containerannotation = root.findtext("containerannotation")

    def has_dep(self, name):
        for i in self.deps:
            if i.name == name:
                return True
        return False

    def remove_dep(self, name):
        # we need to iterate over all deps because if this a
        # kiwi build the same package might appear multiple times
        # NOTE: do not loop and remove items, the second same one would not get catched
        self.deps = [i for i in self.deps if not i.name == name]


class Pac:
    """represent a package to be downloaded

    We build a map that's later used to fill our URL templates
    """

    def __init__(self, node, buildarch, pacsuffix, apiurl, localpkgs=None):
        localpkgs = localpkgs or []

        # set attributes to mute pylint error E1101: Instance of 'Pac' has no '<attr>' member (no-member)
        self.project = None
        self.name = None
        self.canonname = None
        self.repository = None
        self.repoarch = None

        self.mp = {}
        for i in ['binary', 'package',
                  'epoch', 'version', 'release', 'hdrmd5',
                  'project', 'repository', 'sysroot',
                  'preinstall', 'vminstall', 'runscripts',
                  'noinstall', 'installonly', 'notmeta',
                  ]:
            self.mp[i] = node.get(i)

        self.mp['buildarch'] = buildarch
        self.mp['pacsuffix'] = pacsuffix

        self.mp['arch'] = node.get('arch') or self.mp['buildarch']
        self.mp['name'] = node.get('name') or self.mp['binary']

        # this is not the ideal place to check if the package is a localdep or not
        localdep = self.mp['name'] in localpkgs  # and not self.mp['noinstall']
        if not localdep and not (node.get('project') and node.get('repository')):
            raise oscerr.APIError('incomplete information for package %s, may be caused by a broken project configuration.'
                                  % self.mp['name'])

        if not localdep:
            self.mp['extproject'] = node.get('project').replace(':', ':/')
            self.mp['extrepository'] = node.get('repository').replace(':', ':/')
        self.mp['repopackage'] = node.get('package') or '_repository'
        self.mp['repoarch'] = node.get('repoarch') or self.mp['buildarch']

        if pacsuffix == 'deb' and not (self.mp['name'] and self.mp['arch'] and self.mp['version']):
            raise oscerr.APIError(
                "buildinfo for package %s/%s/%s is incomplete"
                % (self.mp['name'], self.mp['arch'], self.mp['version']))

        self.mp['apiurl'] = apiurl

        if self.mp['epoch'] is None:
            epoch = None
        else:
            epoch = self.mp['epoch'].encode()

        if self.mp['release'] is None:
            release = None
        else:
            release = self.mp['release'].encode()

        if self.mp['binary'] == 'updateinfo.xml':
            canonname = 'updateinfo.xml'
        elif self.mp['name'].startswith('container:'):
            canonname = self.mp['name'] + '.tar.xz'
        elif pacsuffix == 'deb':
            canonname = debquery.DebQuery.filename(self.mp['name'].encode(), epoch, self.mp['version'].encode(), release, self.mp['arch'].encode())
        elif pacsuffix == 'arch':
            canonname = archquery.ArchQuery.filename(self.mp['name'].encode(), epoch, self.mp['version'].encode(), release, self.mp['arch'].encode())
        else:
            canonname = rpmquery.RpmQuery.filename(self.mp['name'].encode(), epoch, self.mp['version'].encode(), release or b'0', self.mp['arch'].encode())

        self.mp['canonname'] = decode_it(canonname)
        # maybe we should rename filename key to binary
        self.mp['filename'] = node.get('binary') or decode_it(canonname)
        if self.mp['repopackage'] == '_repository':
            self.mp['repofilename'] = self.mp['name']
        else:
            # OBS 2.3 puts binary into product bdeps (noinstall ones)
            self.mp['repofilename'] = self.mp['filename']

        # make the content of the dictionary accessible as class attributes
        self.__dict__.update(self.mp)

    def makeurls(self, cachedir, urllist):
        self.localdir = '%s/%s/%s/%s' % (cachedir, self.project, self.repository, self.repoarch)
        self.fullfilename = os.path.join(self.localdir, self.canonname)
        self.urllist = [url % self.mp for url in urllist]

    def __str__(self):
        return self.name or ""

    def __repr__(self):
        return "%s" % self.name


def get_preinstall_image(apiurl, arch, cache_dir, img_info, offline=False):
    """
    Searches preinstall image according to build info and downloads it to cache
    (unless offline is set to ``True`` (default: ``False``)).
    Returns preinstall image path, source and list of image binaries, which can
    be used to create rpmlist.

    .. note::
        preinstall image can be used only for new build roots!
    """
    imagefile = ''
    imagesource = ''
    img_bins = []
    for bin in img_info.findall('binary'):
        img_bins.append(bin.text)

    img_project = img_info.get('project')
    img_repository = img_info.get('repository')
    img_arch = arch
    img_pkg = img_info.get('package')
    img_file = img_info.get('filename')
    img_hdrmd5 = img_info.get('hdrmd5')
    if not img_hdrmd5:
        img_hdrmd5 = img_file
    cache_path = '%s/%s/%s/%s' % (cache_dir, img_project, img_repository, img_arch)
    ifile_path = '%s/%s' % (cache_path, img_file)
    ifile_path_part = '%s.part' % ifile_path

    imagefile = ifile_path
    imagesource = "%s/%s/%s [%s]" % (img_project, img_repository, img_pkg, img_hdrmd5)

    if not os.path.exists(ifile_path):
        if offline:
            return '', '', []
        url = "%s/build/%s/%s/%s/%s/%s" % (apiurl, img_project, img_repository, img_arch, img_pkg, img_file)
        print("downloading preinstall image %s" % imagesource)
        if not os.path.exists(cache_path):
            try:
                os.makedirs(cache_path, mode=0o755)
            except OSError as e:
                print('packagecachedir is not writable for you?', file=sys.stderr)
                print(e, file=sys.stderr)
                sys.exit(1)
        progress_obj = None
        if sys.stdout.isatty():
            progress_obj = create_text_meter(use_pb_fallback=False)
        gr = OscFileGrabber(progress_obj=progress_obj)
        try:
            gr.urlgrab(url, filename=ifile_path_part, text='fetching image')
        except HTTPError as e:
            print("Failed to download! ecode:%i reason:%s" % (e.code, e.reason))
            return ('', '', [])
        # download ok, rename partial file to final file name
        os.rename(ifile_path_part, ifile_path)
    return (imagefile, imagesource, img_bins)


def get_built_files(pacdir, buildtype):
    if buildtype == 'spec':
        debs_dir = os.path.join(pacdir, 'DEBS')
        sdebs_dir = os.path.join(pacdir, 'SDEBS')
        if os.path.isdir(debs_dir) or os.path.isdir(sdebs_dir):
            # (S)DEBS directories detected, list their *.(s)deb files
            b_built = subprocess.Popen(['find', debs_dir, '-name', '*.deb'],
                                       stdout=subprocess.PIPE).stdout.read().strip()
            s_built = subprocess.Popen(['find', sdebs_dir, '-name', '*.sdeb'],
                                       stdout=subprocess.PIPE).stdout.read().strip()
        else:
            # default: (S)RPMS directories and their *.rpm files
            b_built = subprocess.Popen(['find', os.path.join(pacdir, 'RPMS'),
                                        '-name', '*.rpm'],
                                       stdout=subprocess.PIPE).stdout.read().strip()
            s_built = subprocess.Popen(['find', os.path.join(pacdir, 'SRPMS'),
                                        '-name', '*.rpm'],
                                       stdout=subprocess.PIPE).stdout.read().strip()
    elif buildtype == 'kiwi':
        b_built = subprocess.Popen(['find', os.path.join(pacdir, 'KIWI'),
                                    '-type', 'f'],
                                   stdout=subprocess.PIPE).stdout.read().strip()
        s_built = ''
    elif buildtype == 'docker':
        b_built = subprocess.Popen(['find', os.path.join(pacdir, 'DOCKER'),
                                    '-type', 'f'],
                                   stdout=subprocess.PIPE).stdout.read().strip()
        s_built = ''
    elif buildtype == 'podman':
        b_built = subprocess.Popen(['find', os.path.join(pacdir, 'DOCKER'),
                                    '-type', 'f'],
                                   stdout=subprocess.PIPE).stdout.read().strip()
        s_built = ''
    elif buildtype == 'fissile':
        b_built = subprocess.Popen(['find', os.path.join(pacdir, 'FISSILE'),
                                    '-type', 'f'],
                                   stdout=subprocess.PIPE).stdout.read().strip()
        s_built = ''
    elif buildtype in ('dsc', 'collax'):
        b_built = subprocess.Popen(['find', os.path.join(pacdir, 'DEBS'),
                                    '-name', '*.deb'],
                                   stdout=subprocess.PIPE).stdout.read().strip()
        s_built = subprocess.Popen(['find', os.path.join(pacdir, 'SOURCES.DEB'),
                                    '-type', 'f'],
                                   stdout=subprocess.PIPE).stdout.read().strip()
    elif buildtype == 'arch':
        b_built = subprocess.Popen(['find', os.path.join(pacdir, 'ARCHPKGS'),
                                    '-name', '*.pkg.tar*'],
                                   stdout=subprocess.PIPE).stdout.read().strip()
        s_built = ''
    elif buildtype == 'livebuild':
        b_built = subprocess.Popen(['find', os.path.join(pacdir, 'OTHER'),
                                    '-name', '*.iso*'],
                                   stdout=subprocess.PIPE).stdout.read().strip()
        s_built = ''
    elif buildtype == 'helm':
        b_built = subprocess.Popen(['find', os.path.join(pacdir, 'HELM'),
                                    '-type', 'f'],
                                   stdout=subprocess.PIPE).stdout.read().strip()
        s_built = ''
    elif buildtype == 'snapcraft':
        b_built = subprocess.Popen(['find', os.path.join(pacdir, 'OTHER'),
                                    '-name', '*.snap'],
                                   stdout=subprocess.PIPE).stdout.read().strip()
        s_built = ''
    elif buildtype == 'appimage':
        b_built = subprocess.Popen(['find', os.path.join(pacdir, 'OTHER'),
                                    '-name', '*.AppImage'],
                                   stdout=subprocess.PIPE).stdout.read().strip()
        s_built = ''
    elif buildtype == 'simpleimage':
        b_built = subprocess.Popen(['find', os.path.join(pacdir, 'OTHER'),
                                    '-type', 'f'],
                                   stdout=subprocess.PIPE).stdout.read().strip()
        s_built = ''
    elif buildtype == 'flatpak':
        b_built = subprocess.Popen(['find', os.path.join(pacdir, 'OTHER'),
                                    '-type', 'f'],
                                   stdout=subprocess.PIPE).stdout.read().strip()
        s_built = ''
    elif buildtype == 'preinstallimage':
        b_built = subprocess.Popen(['find', os.path.join(pacdir, 'OTHER'),
                                    '-type', 'f'],
                                   stdout=subprocess.PIPE).stdout.read().strip()
        s_built = ''
    elif buildtype == 'productcompose':
        b_built = subprocess.Popen(['find', os.path.join(pacdir, 'PRODUCT'),
                                    '-type', 'f'],
                                   stdout=subprocess.PIPE).stdout.read().strip()
        s_built = ''
    elif buildtype == 'mkosi':
        b_built = subprocess.Popen(['find', os.path.join(pacdir, 'OTHER'),
                                    '-type', 'f'],
                                   stdout=subprocess.PIPE).stdout.read().strip()
        s_built = ''
    else:
        print('WARNING: Unknown package type \'%s\'.' % buildtype, file=sys.stderr)
        b_built = ''
        s_built = ''
    return s_built, b_built


def get_repo(path):
    """
    Walks up path looking for any repodata directories.

    :param path: path to a directory
    :return: path to repository directory containing repodata directory with repomd.xml file
    :rtype: str
    """

    for root, dirs, files in os.walk(path):
        if not "repodata" in dirs:
            continue
        if "repomd.xml" in os.listdir(os.path.join(root, "repodata")):
            return root
    return None


def get_prefer_pkgs(dirs, wanted_arch, type, cpio):
    paths = []
    repositories = []

    suffix = '*.rpm'
    if type in ('dsc', 'collax', 'livebuild'):
        suffix = '*.deb'
    elif type == 'arch':
        suffix = '*.pkg.tar.*'

    for dir in dirs:
        # check for repodata
        repository = get_repo(dir)
        if repository is None:
            paths += glob.glob(os.path.join(os.path.abspath(dir), suffix))
        else:
            repositories.append(repository)

    packageQueries = packagequery.PackageQueries(wanted_arch)

    for repository in repositories:
        repodataPackageQueries = repodata.queries(repository)

        for packageQuery in repodataPackageQueries:
            packageQueries.add(packageQuery)

    for path in paths:
        if path.endswith('.src.rpm') or path.endswith('.nosrc.rpm'):
            continue
        if path.endswith('.patch.rpm') or path.endswith('.delta.rpm'):
            continue
        packageQuery = packagequery.PackageQuery.query(path)
        packageQueries.add(packageQuery)

    prefer_pkgs = {decode_it(name): packageQuery.path()
                   for name, packageQuery in packageQueries.items()}

    depfile = create_deps(packageQueries.values())
    cpio.add(b'deps', b'\n'.join(depfile))
    return prefer_pkgs


def create_deps(pkgqs):
    """
    creates a list of dependencies which corresponds to build's internal
    dependency file format
    """
    depfile = []
    for p in pkgqs:
        id = b'%s.%s-0/0/0: ' % (p.name(), p.arch())
        depfile.append(b'P:%s%s' % (id, b' '.join(p.provides())))
        depfile.append(b'R:%s%s' % (id, b' '.join(p.requires())))
        d = p.conflicts()
        if d:
            depfile.append(b'C:%s%s' % (id, b' '.join(d)))
        d = p.obsoletes()
        if d:
            depfile.append(b'O:%s%s' % (id, b' '.join(d)))
        d = p.recommends()
        if d:
            depfile.append(b'r:%s%s' % (id, b' '.join(d)))
        d = p.supplements()
        if d:
            depfile.append(b's:%s%s' % (id, b' '.join(d)))
        depfile.append(b'I:%s%s-%s 0-%s' % (id, p.name(), p.evr(), p.arch()))
    return depfile


trustprompt = """Would you like to ...
0 - quit (default)
1 - always trust packages from '%(project)s'
2 - trust packages just this time
? """


def check_trusted_projects(apiurl, projects, interactive=True):
    trusted = conf.config['api_host_options'][apiurl]['trusted_prj']
    tlen = len(trusted)
    for prj in projects:
        is_trusted = False
        for pattern in trusted:
            if fnmatch.fnmatch(prj, pattern):
                is_trusted = True
                break
        if not is_trusted:
            print("\nThe build root needs packages from project '%s'." % prj)
            print("Note that malicious packages can compromise the build result or even your system.")

            if interactive:
                r = raw_input(trustprompt % {'project': prj})
            else:
                r = "0"

            if r == '1':
                print("adding '%s' to oscrc: ['%s']['trusted_prj']" % (prj, apiurl))
                trusted.append(prj)
            elif r != '2':
                print("Well, goodbye then :-)")
                raise oscerr.UserAbort()

    if tlen != len(trusted):
        conf.config['api_host_options'][apiurl]['trusted_prj'] = trusted
        conf.config_set_option(apiurl, 'trusted_prj', ' '.join(trusted))


def get_kiwipath_from_buildinfo(bi, prj, repo):
    # If the project does not have a path defined we need to get the config
    # via the repositories in the kiwi file. Unfortunately the buildinfo
    # does not include a hint if this is the case, so we rely on a heuristic
    # here: if the path list contains our own repo, it probably does not
    # come from the kiwi file and thus a path is defined in the config.
    # It is unlikely that our own repo is included in the kiwi file, as it
    # contains no packages.
    myprp = prj + '/' + repo
    if myprp in bi.pathes:
        return None
    kiwipath = bi.pathes
    kiwipath.insert(0, myprp)
    return kiwipath


def calculate_prj_pac(store, opts, descr):
    project = opts.alternative_project or store.project
    if opts.local_package:
        package = os.path.splitext(os.path.basename(descr))[0]
    else:
        store.assert_is_package()
        package = store.package
    return project, package


def calculate_build_root_user(vm_type):
    if vm_type in ("kvm", "podman", "qemu"):
        return getpass.getuser()
    return None


def calculate_build_root(apihost, prj, pac, repo, arch, user=None):
    user = user or ""
    dash_user = f"-{user:s}" if user else ""
    buildroot = conf.config["build-root"] % {
        'apihost': apihost,
        'project': prj,
        'package': pac,
        'repo': repo,
        'arch': arch,
        "user": user,
        "dash_user": dash_user,
    }
    return buildroot


def build_as_user(vm_type=None):
    if not conf.config.su_wrapper:
        return True
    if calculate_build_root_user(vm_type):
        return True
    return False


def su_wrapper(cmd):
    sucmd = conf.config['su-wrapper'].split()
    if sucmd:
        if sucmd[0] == 'su':
            if sucmd[-1] == '-c':
                sucmd.pop()
            cmd = sucmd + ['-s', cmd[0], 'root', '--'] + cmd[1:]
        else:
            cmd = sucmd + cmd
    return cmd


def run_build(opts, *args):
    cmd = [conf.config['build-cmd']]
    cmd += args

    if opts.vm_type:
        cmd.extend(["--vm-type", opts.vm_type])

    user = calculate_build_root_user(opts.vm_type)
    if not user:
        cmd = su_wrapper(cmd)

    if not opts.userootforbuild:
        cmd.append('--norootforbuild')
    return run_external(cmd[0], *cmd[1:])


def create_build_descr_data(
    build_descr_path: Optional[str],
    *,
    build_type: Optional[str],
    repo: Optional[str] = None,
    arch: Optional[str] = None,
    prefer_pkgs: Optional[List[str]] = None,
    define: Optional[List[str]] = None,
    define_with: Optional[List[str]] = None,
    define_without: Optional[List[str]] = None,
):
    if build_descr_path:
        build_descr_path = os.path.abspath(build_descr_path)
        topdir = os.path.dirname(build_descr_path)
    else:
        topdir = None
    result_data = []

    if build_descr_path:
        print(f"Using local file: {os.path.basename(build_descr_path)}", file=sys.stderr)
        with open(build_descr_path, "rb") as f:
            build_descr_data = f.read()

        # HACK: there's no api to provide custom defines
        # TODO: check if we're working with a spec?
        defines: List[bytes] = []
        for i in define or []:
            defines.append(f"%define {i}".encode("utf-8"))
        for i in define_with or []:
            defines.append(f"%define _with_{i} 1".encode("utf-8"))
        for i in define_without or []:
            defines.append(f"%define _without_{i} 1".encode("utf-8"))
        if defines:
            build_descr_data = b"\n".join(defines) + b"\n\n" + build_descr_data

        # build recipe must go first for compatibility with the older OBS versions
        result_data.append((os.path.basename(build_descr_path).encode("utf-8"), build_descr_data))

    if topdir:
        for include_file in glob.glob(os.path.join(topdir, "*.inc")):
            fn = os.path.basename(include_file)
            print(f"Using local file: {fn}", file=sys.stderr)
            with open(include_file, "rb") as f:
                result_data.append((fn.encode("utf-8"), f.read()))

    if topdir:
        buildenv_file = os.path.join(topdir, f"_buildenv.{repo}.{arch}")
        if not os.path.isfile(buildenv_file):
            buildenv_file = os.path.join(topdir, "_buildenv")
        if os.path.isfile(buildenv_file):
            print(f"Using local file: {os.path.basename(buildenv_file)}", file=sys.stderr)
            with open(buildenv_file, "rb") as f:
                result_data.append((b"buildenv", f.read()))

    if topdir:
        service_file = os.path.join(topdir, "_service")
        if os.path.isfile(service_file):
            print("Using local file: _service", file=sys.stderr)
            with open(service_file, "rb") as f:
                result_data.append((b"_service", f.read()))

    if not result_data and not prefer_pkgs:
        return None, {}

    cpio_data = cpio.CpioWrite()
    for key, value in result_data:
        cpio_data.add(key, value)

    if prefer_pkgs:
        print(f"Scanning the following dirs for local preferred packages: {', '.join(prefer_pkgs)}", file=sys.stderr)
        prefer_pkgs_result = get_prefer_pkgs(prefer_pkgs, arch, build_type, cpio_data)
    else:
        prefer_pkgs_result = {}

    return cpio_data.get(), prefer_pkgs_result


def main(apiurl, store, opts, argv):

    repo = argv[0]
    arch = argv[1]
    build_descr = argv[2]
    xp = []
    build_root = None
    cache_dir = None
    build_uid = ''
    config = conf.config
    build_shell_after_fail = config['build-shell-after-fail']
    vm_memory = config['build-memory']
    vm_disk_size = config['build-vmdisk-rootsize']
    vm_type = config['build-type']
    vm_telnet = None

    build_descr = os.path.abspath(build_descr)
    build_type = os.path.splitext(build_descr)[1][1:]
    if build_type in ['spec', 'dsc', 'kiwi', 'productcompose', 'livebuild']:
        # File extension works
        pass
    elif os.path.basename(build_descr) == 'PKGBUILD':
        build_type = 'arch'
    elif os.path.basename(build_descr) == 'build.collax':
        build_type = 'collax'
    elif os.path.basename(build_descr) == 'appimage.yml':
        build_type = 'appimage'
    elif os.path.basename(build_descr) == 'Chart.yaml':
        build_type = 'helm'
    elif os.path.basename(build_descr) == 'snapcraft.yaml':
        build_type = 'snapcraft'
    elif os.path.basename(build_descr) == 'simpleimage':
        build_type = 'simpleimage'
    elif os.path.basename(build_descr) == 'Containerfile' or os.path.basename(build_descr).startswith('Containerfile.'):
        build_type = 'docker'
    elif os.path.basename(build_descr) == 'Dockerfile' or os.path.basename(build_descr).startswith('Dockerfile.'):
        build_type = 'docker'
    elif os.path.basename(build_descr) == 'fissile.yml':
        build_type = 'fissile'
    elif os.path.basename(build_descr) == '_preinstallimage':
        build_type = 'preinstallimage'
    elif build_descr.endswith('flatpak.yaml') or build_descr.endswith('flatpak.yml') or build_descr.endswith('flatpak.json'):
        build_type = 'flatpak'
    elif os.path.basename(build_descr).startswith('mkosi.'):
        build_type = 'mkosi'
    else:
        raise oscerr.WrongArgs(
            'Unknown build type: \'%s\'. '
            'Build description should end in .spec, .dsc, .kiwi, .productcompose or .livebuild. '
            'Or being named PKGBUILD, build.collax, simpleimage, appimage.yml, '
            'Chart.yaml, snapcraft.yaml, flatpak.json, flatpak.yml, flatpak.yaml, '
            'preinstallimage, Dockerfile.*, Containerfile.* or mkosi.*' % build_type)

    if not os.path.isfile(build_descr):
        raise oscerr.WrongArgs('Error: build description file named \'%s\' does not exist.' % build_descr)

    buildargs = []
    buildargs.append('--statistics')
    if not opts.userootforbuild:
        buildargs.append('--norootforbuild')
    if opts.clean:
        buildargs.append('--clean')
    if opts.checks:
        buildargs.append('--checks')
    if opts.nochecks:
        buildargs.append('--no-checks')
    if not opts.no_changelog:
        buildargs.append('--changelog')
    if opts.root:
        build_root = opts.root
    if opts.target:
        buildargs.append('--target=%s' % opts.target)
    if opts.threads:
        buildargs.append('--threads=%s' % opts.threads)
    if opts.jobs:
        buildargs.append('--jobs=%s' % opts.jobs)
    elif config['build-jobs'] > 0:
        buildargs.append('--jobs=%s' % config['build-jobs'])
    if opts.icecream or config['icecream'] != '0':
        if opts.icecream:
            num = opts.icecream
        else:
            num = config['icecream']

        if int(num) > 0:
            buildargs.append('--icecream=%s' % num)
            xp.append('icecream')
            xp.append('gcc-c++')
    if opts.ccache or config['ccache']:
        buildargs.append('--ccache')
        xp.append('ccache')
    if opts.pkg_ccache:
        buildargs.append('--pkg-ccache=%s' % opts.pkg_ccache)
        xp.append('ccache')
    if opts.linksources:
        buildargs.append('--linksources')
    if opts.baselibs:
        buildargs.append('--baselibs')
    if opts.debuginfo:
        buildargs.append('--debug')
    if opts._with:
        for o in opts._with:
            buildargs.append('--with=%s' % o)
    if opts.without:
        for o in opts.without:
            buildargs.append('--without=%s' % o)
    if opts.define:
        for o in opts.define:
            buildargs.append('--define=%s' % o)
    if config['build-uid']:
        build_uid = config['build-uid']
    if opts.build_uid:
        build_uid = opts.build_uid
    if build_uid:
        buildidre = re.compile('^[0-9]+:[0-9]+$')
        if build_uid == 'caller':
            buildargs.append('--uid=%s:%s' % (os.getuid(), os.getgid()))
        elif buildidre.match(build_uid):
            buildargs.append('--uid=%s' % build_uid)
        else:
            print('Error: build-uid arg must be 2 colon separated numerics: "uid:gid" or "caller"', file=sys.stderr)
            return 1
    if opts.shell_after_fail:
        build_shell_after_fail = opts.shell_after_fail
    if opts.vm_memory:
        vm_memory = opts.vm_memory
    if opts.vm_disk_size:
        vm_disk_size = opts.vm_disk_size
    if opts.vm_type:
        vm_type = opts.vm_type
    if opts.vm_telnet:
        vm_telnet = opts.vm_telnet
    if opts.alternative_project:
        prj = opts.alternative_project
        pac = '_repository'
    else:
        prj = store.project
        if opts.local_package:
            pac = '_repository'
        else:
            pac = store.package
    if opts.multibuild_package:
        buildargs.append('--buildflavor=%s' % opts.multibuild_package)
        pac = pac + ":" + opts.multibuild_package
    if opts.verbose_mode:
        buildargs.append('--verbose=%s' % opts.verbose_mode)
    if opts.no_timestamps:
        buildargs.append('--no-timestamps')
    if opts.wipe:
        buildargs.append("--wipe")

    pacname = pac
    if pacname == '_repository':
        if not opts.local_package:
            try:
                pacname = store.package
            except oscerr.NoWorkingCopy:
                opts.local_package = True
        if opts.local_package:
            pacname = os.path.splitext(os.path.basename(build_descr))[0]
    apihost = urlsplit(apiurl)[1]
    if not build_root:
        user = calculate_build_root_user(vm_type)
        build_root = calculate_build_root(apihost, prj, pacname, repo, arch, user)

    # We configure sccache after pacname, so that in default cases we can have an sccache for each
    # package to prevent cross-cache polutions. It helps to make the local-use case a bit nicer.
    if opts.sccache_uri or config['sccache_uri'] or opts.sccache or config['sccache']:
        if opts.pkg_ccache or opts.ccache or config['ccache']:
            raise oscerr.WrongArgs('Error: sccache and ccache can not be enabled at the same time')
        sccache_arg = "--sccache-uri=/var/tmp/osbuild-sccache-{pkgname}.tar"
        if opts.sccache_uri:
            sccache_arg = '--sccache-uri=%s' % opts.sccache_uri
        elif config['sccache_uri']:
            sccache_arg = '--sccache-uri=%s' % config['sccache_uri']
        # Format the package name.
        sccache_arg = sccache_arg.format(pkgname=pacname)
        buildargs.append(sccache_arg)
        xp.append('sccache')

    # define buildinfo & config local cache
    bi_file = None
    bc_file = None
    bi_filename = '_buildinfo-%s-%s.xml' % (repo, arch)
    bc_filename = '_buildconfig-%s-%s' % (repo, arch)
    if store is not None and store.is_package and os.access(core.store, os.W_OK):
        bi_filename = os.path.join(os.getcwd(), core.store, bi_filename)
        bc_filename = os.path.join(os.getcwd(), core.store, bc_filename)
    elif not os.access('.', os.W_OK):
        bi_file = NamedTemporaryFile(prefix=bi_filename)
        bi_filename = bi_file.name
        bc_file = NamedTemporaryFile(prefix=bc_filename)
        bc_filename = bc_file.name
    else:
        bi_filename = os.path.abspath(bi_filename)
        bc_filename = os.path.abspath(bc_filename)

    if opts.shell:
        buildargs.append("--shell")

    if build_shell_after_fail:
        buildargs.append("--shell-after-fail")

    if opts.shell_cmd:
        buildargs.append("--shell-cmd")
        buildargs.append(opts.shell_cmd)

    if opts.noinit:
        buildargs.append('--noinit')

    if opts.local_package or not store.is_package:
        opts.skip_local_service_run = True

    # check for source services
    if not opts.offline and not opts.skip_local_service_run:
        p = core.Package(os.curdir)
        r = p.run_source_services(verbose=True)
        if r:
            raise oscerr.ServiceRuntimeError('Source service run failed!')

    cache_dir = config['packagecachedir'] % {'apihost': apihost}

    extra_pkgs = []
    if not opts.extra_pkgs:
        extra_pkgs = config.get('extra-pkgs', [])
    elif opts.extra_pkgs != ['']:
        extra_pkgs = opts.extra_pkgs

    if opts.extra_pkgs_from:
        for filename in opts.extra_pkgs_from:
            with open(filename, encoding="utf-8") as f:
                for line in f:
                    extra_pkgs.append(line.rstrip('\n'))

    if xp:
        extra_pkgs += xp

    build_descr_data, prefer_pkgs = create_build_descr_data(
        build_descr,
        build_type=build_type,
        repo=repo,
        arch=arch,
        prefer_pkgs=opts.prefer_pkgs,
        define=opts.define,
        define_with=opts._with,
        define_without=opts.without,
    )

    # special handling for overlay and rsync-src/dest
    specialcmdopts = []
    if opts.rsyncsrc or opts.rsyncdest:
        if not opts.rsyncsrc or not opts.rsyncdest:
            raise oscerr.WrongOptions('When using --rsync-{src,dest} both parameters have to be specified.')
        myrsyncsrc = os.path.abspath(os.path.expanduser(os.path.expandvars(opts.rsyncsrc)))
        if not os.path.isdir(myrsyncsrc):
            raise oscerr.WrongOptions('--rsync-src %s is no valid directory!' % opts.rsyncsrc)
        # can't check destination - its in the target chroot ;) - but we can check for sanity
        myrsyncdest = os.path.expandvars(opts.rsyncdest)
        if not os.path.isabs(myrsyncdest):
            raise oscerr.WrongOptions('--rsync-dest %s is no absolute path (starting with \'/\')!' % opts.rsyncdest)
        specialcmdopts = ['--rsync-src=' + myrsyncsrc, '--rsync-dest=' + myrsyncdest]
    if opts.overlay:
        myoverlay = os.path.abspath(os.path.expanduser(os.path.expandvars(opts.overlay)))
        if not os.path.isdir(myoverlay):
            raise oscerr.WrongOptions('--overlay %s is no valid directory!' % opts.overlay)
        specialcmdopts += ['--overlay=' + myoverlay]

    try:
        if opts.noinit:
            if not os.path.isfile(bi_filename):
                raise oscerr.WrongOptions('--noinit is not possible, no local buildinfo file')
            print('Use local \'%s\' file as buildinfo' % bi_filename)
            if not os.path.isfile(bc_filename):
                raise oscerr.WrongOptions('--noinit is not possible, no local buildconfig file')
            print('Use local \'%s\' file as buildconfig' % bc_filename)
        elif opts.offline:
            if not os.path.isfile(bi_filename):
                raise oscerr.WrongOptions('--offline is not possible, no local buildinfo file')
            print('Use local \'%s\' file as buildinfo' % bi_filename)
            if not os.path.isfile(bc_filename):
                raise oscerr.WrongOptions('--offline is not possible, no local buildconfig file')
        else:
            print('Getting buildconfig from server and store to %s' % bc_filename)
            bc = get_buildconfig(apiurl, prj, repo)
            if not bc_file:
                bc_file = open(bc_filename, 'w')
            bc_file.write(decode_it(bc))
            bc_file.flush()
            if os.path.exists('/usr/lib/build/queryconfig') and not opts.nodebugpackages:
                debug_pkgs = decode_it(return_external('/usr/lib/build/queryconfig', '--dist', bc_filename, 'substitute', 'obs:cli_debug_packages'))
                if len(debug_pkgs) > 0:
                    extra_pkgs.extend(debug_pkgs.strip().split(" "))

            print('Getting buildinfo from server and store to %s' % bi_filename)
            bi_text = decode_it(get_buildinfo(apiurl,
                                              prj,
                                              pac,
                                              repo,
                                              arch,
                                              specfile=build_descr_data,
                                              addlist=extra_pkgs))
            if not bi_file:
                bi_file = open(bi_filename, 'w')
            # maybe we should check for errors before saving the file
            bi_file.write(bi_text)
            bi_file.flush()
            kiwipath = None
            if build_type == 'kiwi':
                bi = Buildinfo(bi_filename, apiurl, 'kiwi', list(prefer_pkgs.keys()))
                kiwipath = get_kiwipath_from_buildinfo(bi, prj, repo)
                bc = get_buildconfig(apiurl, prj, repo, kiwipath)
                bc_file.seek(0)
                bc_file.write(decode_it(bc))
                bc_file.flush()
    except HTTPError as e:
        if e.code == 404:
            # check what caused the 404
            if meta_exists(metatype='prj', path_args=(prj, ),
                           template_args=None, create_new=False, apiurl=apiurl):
                pkg_meta_e = None
                try:
                    # take care, not to run into double trouble.
                    pkg_meta_e = meta_exists(metatype='pkg', path_args=(prj, pac),
                                             template_args=None, create_new=False,
                                             apiurl=apiurl)
                except:
                    pass

                if pkg_meta_e:
                    print('ERROR: Either wrong repo/arch as parameter or a parse error of .spec/.dsc/.kiwi file due to syntax error', file=sys.stderr)
                else:
                    print('The package \'%s\' does not exist - please '
                          'rerun with \'--local-package\'' % pac, file=sys.stderr)
            else:
                print('The project \'%s\' does not exist - please '
                      'rerun with \'--alternative-project <alternative_project>\'' % prj, file=sys.stderr)
            sys.exit(1)
        else:
            raise

    # Set default binary type if cannot be detected
    binary_type = 'rpm'
    if os.path.exists('/usr/lib/build/queryconfig'):
        binary_type = decode_it(return_external('/usr/lib/build/queryconfig', '--dist', bc_filename, 'binarytype')).strip()
    # If binary type is set to a useless value, reset to 'rpm'
    if binary_type == 'UNDEFINED':
        binary_type = 'rpm'

    bi = Buildinfo(bi_filename, apiurl, build_type, list(prefer_pkgs.keys()), binary_type)

    if bi.debuginfo and not (opts.disable_debuginfo or '--debug' in buildargs):
        buildargs.append('--debug')

    if opts.release:
        bi.release = opts.release

    if bi.release:
        buildargs.append('--release')
        buildargs.append(bi.release)

    if opts.stage:
        buildargs.append('--stage')
        buildargs.append(opts.stage)

    if opts.build_opt:
        buildargs += opts.build_opt

    if opts.buildtool_opt:
        buildargs += [f"--buildtool-opt={opt}" for opt in opts.buildtool_opt]

    # real arch of this machine
    # vs.
    # arch we are supposed to build for
    if vm_type != "emulator" and vm_type != "qemu":
        if bi.hostarch is not None:
            if hostarch != bi.hostarch and bi.hostarch not in can_also_build.get(hostarch, []):
                print('Error: hostarch \'%s\' is required.' % (bi.hostarch), file=sys.stderr)
                return 1
        elif hostarch != bi.buildarch:
            if bi.buildarch not in can_also_build.get(hostarch, []):
                print('WARNING: It is guessed to build on hostarch \'%s\' for \'%s\' via QEMU user emulation.' % (hostarch, bi.buildarch), file=sys.stderr)

    rpmlist_prefers = []
    if prefer_pkgs:
        print('Evaluating preferred packages')
        for name, path in prefer_pkgs.items():
            if bi.has_dep(name):
                # We remove a preferred package from the buildinfo, so that the
                # fetcher doesn't take care about them.
                # Instead, we put it in a list which is appended to the rpmlist later.
                # At the same time, this will make sure that these packages are
                # not verified.
                bi.remove_dep(name)
                rpmlist_prefers.append((name, path))
                print(' - %s (%s)' % (name, path))

    print('Updating cache of required packages')

    urllist = []
    if not opts.download_api_only:
        # transform 'url1, url2, url3' form into a list
        if 'urllist' in config:
            if isinstance(config['urllist'], str):
                re_clist = re.compile('[, ]+')
                urllist = [i.strip() for i in re_clist.split(config['urllist'].strip())]
            else:
                urllist = config['urllist']

        # OBS 1.5 and before has no downloadurl defined in buildinfo, but it is obsolete again meanwhile.
        # we have now specific download repositories per repository. Could be removed IMHO, since the api fallback
        # is there. In worst case it could fetch the wrong rpm...
        if bi.downloadurl:
            urllist.append(bi.downloadurl.replace('%', '%%') + '/%(extproject)s/%(extrepository)s/%(arch)s/%(filename)s')
    if opts.disable_cpio_bulk_download:
        urllist.append('%(apiurl)s/build/%(project)s/%(repository)s/%(repoarch)s/%(repopackage)s/%(repofilename)s')

    fetcher = Fetcher(cache_dir,
                      urllist=urllist,
                      offline=opts.noinit or opts.offline,
                      http_debug=config['http_debug'],
                      modules=bi.modules,
                      enable_cpio=not opts.disable_cpio_bulk_download and bi.enable_cpio,
                      cookiejar=connection.CookieJarAuthHandler(apiurl, os.path.expanduser(config["cookiejar"]))._cookiejar,
                      download_api_only=opts.download_api_only)

    if not opts.trust_all_projects:
        # implicitly trust the project we are building for
        check_trusted_projects(apiurl, [i for i in bi.projects.keys() if not i == prj])

    imagefile = ''
    imagesource = ''
    imagebins = []
    if build_as_user(vm_type):
        # preinstallimage extraction will fail because unprivileged user cannot chroot or extract devices from the tarball
        bi.preinstallimage = None
    if build_type == 'preinstallimage':
        # preinstallimage would repackage just the previously built preinstallimage
        bi.preinstallimage = None

    if (not config['no_preinstallimage'] and not opts.nopreinstallimage and
        bi.preinstallimage and
        not opts.noinit and
        (opts.clean or (not os.path.exists(build_root + "/installed-pkg") and
                        not os.path.exists(build_root + "/.build/init_buildsystem.data")))):
        (imagefile, imagesource, imagebins) = get_preinstall_image(apiurl, arch, cache_dir, bi.preinstallimage,
                                                                   opts.offline)
        if imagefile:
            # remove binaries from build deps which are included in preinstall image
            for i in bi.deps:
                if i.name in imagebins:
                    bi.remove_dep(i.name)

    # now update the package cache
    fetcher.run(bi)

    old_pkg_dir = None
    if opts.oldpackages:
        old_pkg_dir = opts.oldpackages
        if not old_pkg_dir.startswith('/') and not opts.offline:
            data = [prj, pacname, repo, arch]
            if old_pkg_dir == '_link':
                p = core.Package(os.curdir)
                if not p.islink():
                    raise oscerr.WrongOptions('package is not a link')
                data[0] = p.linkinfo.project
                data[1] = p.linkinfo.package
                repos = core.get_repositories_of_project(apiurl, data[0])
                # hack for links to e.g. Factory
                if not data[2] in repos and 'standard' in repos:
                    data[2] = 'standard'
            elif old_pkg_dir != '' and old_pkg_dir != '_self':
                a = old_pkg_dir.split('/')
                for i in range(0, len(a)):
                    data[i] = a[i]

            destdir = os.path.join(cache_dir, data[0], data[2], data[3])
            old_pkg_dir = None
            try:
                print("Downloading previous build from %s ..." % '/'.join(data))
                binaries = get_binarylist(apiurl, data[0], data[2], data[3], package=data[1], verbose=True)
            except Exception as e:
                print("Error: failed to get binaries: %s" % str(e))
                binaries = []

            if binaries:
                class mytmpdir:
                    """ temporary directory that removes itself"""

                    def __init__(self, *args, **kwargs):
                        self.name = mkdtemp(*args, **kwargs)
                    _rmtree = staticmethod(shutil.rmtree)

                    def cleanup(self):
                        self._rmtree(self.name)

                    def __del__(self):
                        self.cleanup()

                    def __exit__(self, exc_type, exc_value, traceback):
                        self.cleanup()

                    def __str__(self):
                        return self.name or ""

                old_pkg_dir = mytmpdir(prefix='.build.oldpackages', dir=os.path.abspath(os.curdir))
                if not os.path.exists(destdir):
                    os.makedirs(destdir)
            for i in binaries:
                fname = os.path.join(destdir, i.name)
                os.symlink(fname, os.path.join(str(old_pkg_dir), i.name))
                if os.path.exists(fname):
                    st = os.stat(fname)
                    if st.st_mtime == i.mtime and st.st_size == i.size:
                        continue
                get_binary_file(apiurl,
                                data[0],
                                data[2], data[3],
                                i.name,
                                package=data[1],
                                target_filename=fname,
                                target_mtime=i.mtime,
                                progress_meter=True)

        if old_pkg_dir is not None:
            buildargs.append('--oldpackages=%s' % old_pkg_dir)

    # Make packages from buildinfo available as repos for kiwi/docker/fissile
    if build_type in ('kiwi', 'docker', 'podman', 'fissile', 'productcompose'):
        if os.path.lexists('repos'):
            shutil.rmtree('repos')
        if os.path.lexists('containers'):
            shutil.rmtree('containers')
        os.mkdir('repos')
        for i in bi.deps:
            if not i.extproject:
                # remove
                bi.deps.remove(i)
                continue
            if i.notmeta:
                continue
            # project
            pdir = str(i.extproject).replace(':/', ':')
            # repo
            rdir = str(i.extrepository).replace(':/', ':')
            # arch
            adir = i.repoarch
            # source fullfilename
            sffn = i.fullfilename
            filename = sffn.split("/")[-1]
            if i.name == 'updateinfo.xml':
                adir = 'updateinfo'
                filename = i.package + ':' + i.repoarch + ':updateinfo.xml'
            # project/repo
            if i.name.startswith("container:"):
                prdir = "containers/" + pdir + "/" + rdir
                pradir = prdir
                filename = filename[10:]
                if build_type == 'kiwi':
                    buildargs.append('--kiwi-parameter')
                    buildargs.append('--set-container-derived-from=dir://./' + prdir + "/" + filename)
            else:
                prdir = "repos/" + pdir + "/" + rdir
                # project/repo/arch
                pradir = prdir + "/" + adir
            # target fullfilename
            tffn = pradir + "/" + filename
            if not os.path.exists(os.path.join(pradir)):
                os.makedirs(os.path.join(pradir))
            if not os.path.exists(tffn):
                print("Using package: " + sffn)
                if opts.linksources:
                    os.link(sffn, tffn)
                else:
                    os.symlink(sffn, tffn)
            if prefer_pkgs:
                for name, path in prefer_pkgs.items():
                    if name == filename:
                        print("Using prefered package: " + path + "/" + filename)
                        os.unlink(tffn)

        if bi.containerannotation:
            if not os.path.exists("containers"):
                os.makedirs("containers")
            with open("containers/annotation", "wb") as f:
                f.write(bi.containerannotation.encode())

        if prefer_pkgs:
            localpkgdir = "repos/_local/"
            os.mkdir(localpkgdir)
            buildargs.append("--kiwi-parameter")
            buildargs.append("--add-repo")
            buildargs.append("--kiwi-parameter")
            buildargs.append(f"dir://./{localpkgdir}")
            buildargs.append("--kiwi-parameter")
            buildargs.append("--add-repotype")
            buildargs.append("--kiwi-parameter")
            buildargs.append("rpm-md")
            for name, path in prefer_pkgs.items():
                tffn = os.path.join(localpkgdir, os.path.basename(path))
                if opts.linksources:
                    os.link(path, tffn)
                else:
                    os.symlink(path, tffn)

    else:
        buildargs.append("--repos-directory=-")
        buildargs.append("--containers-directory=-")

    if build_type == 'kiwi':
        # Is a obsrepositories tag used?
        try:
            tree = xml_parse(build_descr)
        except:
            print('could not parse the kiwi file:', file=sys.stderr)
            print(open(build_descr).read(), file=sys.stderr)
            sys.exit(1)
        root = tree.getroot()

        # product
        if root.find('instsource'):
            # leads to unsigned media, but avoids build failure
            buildargs.append('--signdummy')

        for xml in root.findall('instsource'):
            found_obsrepositories = 0
            for node in xml.findall('instrepo'):
                if node and node.find('source').get('path') == 'obsrepositories:/':
                    for path in bi.pathes:
                        found_obsrepositories += 1
                        new_node = ET.SubElement(xml, 'instrepo')
                        new_node.set('name', node.get('name') + "_" + str(found_obsrepositories))
                        new_node.set('priority', node.get('priority'))
                        new_node.set('local', 'true')
                        new_source_node = ET.SubElement(new_node, 'source')
                        new_source_node.set('path', "obs://" + path)
                    xml.remove(node)

            if found_obsrepositories > 0:
                build_descr = os.getcwd() + '/_service:osc_obsrepositories:' + build_descr.rsplit('/', 1)[-1]
                tree.write(open(build_descr, 'wb'))

        # appliance
        expand_obsrepos = None
        for xml in root.findall('repository'):
            if xml.find('source').get('path') == 'obsrepositories:/':
                expand_obsrepos = True
        if expand_obsrepos:
            buildargs.append('--kiwi-parameter')
            buildargs.append('--ignore-repos')
            for xml in root.findall('repository'):
                if xml.find('source').get('path') == 'obsrepositories:/':
                    for path in bi.pathes:
                        if not os.path.isdir("repos/" + path):
                            continue
                        buildargs.append('--kiwi-parameter')
                        buildargs.append('--add-repo')
                        buildargs.append('--kiwi-parameter')
                        buildargs.append("dir://./repos/" + path)
                        buildargs.append('--kiwi-parameter')
                        buildargs.append('--add-repotype')
                        buildargs.append('--kiwi-parameter')
                        buildargs.append('rpm-md')
                        if xml.get('priority'):
                            buildargs.append('--kiwi-parameter')
                            buildargs.append('--add-repoprio=' + xml.get('priority'))
                else:
                    m = re.match(r"obs://[^/]+/([^/]+)/(\S+)", xml.find('source').get('path'))
                    if not m:
                        # short path without obs instance name
                        m = re.match(r"obs://([^/]+)/(.+)", xml.find('source').get('path'))
                    project = m.group(1).replace(":", ":/")
                    repo = m.group(2)
                    buildargs.append('--kiwi-parameter')
                    buildargs.append('--add-repo')
                    buildargs.append('--kiwi-parameter')
                    buildargs.append("dir://./repos/" + project + "/" + repo)
                    buildargs.append('--kiwi-parameter')
                    buildargs.append('--add-repotype')
                    buildargs.append('--kiwi-parameter')
                    buildargs.append('rpm-md')
                    if xml.get('priority'):
                        buildargs.append('--kiwi-parameter')
                        buildargs.append('--add-repopriority=' + xml.get('priority'))

    if vm_type in ('xen', 'kvm', 'lxc', 'nspawn'):
        print('Skipping verification of package signatures due to secure VM build')
    elif bi.pacsuffix == 'rpm':
        if opts.no_verify:
            print('Skipping verification of package signatures')
        else:
            print('Verifying integrity of cached packages')
            verify_pacs(bi)
    elif bi.pacsuffix == 'deb':
        if opts.no_verify or opts.noinit:
            print('Skipping verification of package signatures')
        else:
            print('WARNING: deb packages get not verified, they can compromise your system !')
    else:
        print('WARNING: unknown packages get not verified, they can compromise your system !')

    for i in bi.deps:
        if i.hdrmd5:
            if not i.name.startswith('container:') and not i.fullfilename.endswith(".rpm"):
                continue
            if i.name.startswith('container:'):
                hdrmd5 = dgst(i.fullfilename)
            else:
                hdrmd5 = packagequery.PackageQuery.queryhdrmd5(i.fullfilename)
            if not hdrmd5:
                print("Error: cannot get hdrmd5 for %s" % i.fullfilename)
                sys.exit(1)
            if hdrmd5 != i.hdrmd5:
                if conf.config["api_host_options"][apiurl]["disable_hdrmd5_check"]:
                    print(f"Warning: Ignoring a hdrmd5 mismatch for {i.fullfilename}: {hdrmd5} (actual) != {i.hdrmd5} (expected)")
                else:
                    print(f"Error: hdrmd5 mismatch for {i.fullfilename}: {hdrmd5} (actual) != {i.hdrmd5} (expected)")
                    sys.exit(1)

    print('Writing build configuration')

    if build_type in ('kiwi', 'docker', 'podman', 'fissile', 'productcompose'):
        rpmlist = ['%s %s\n' % (i.name, i.fullfilename) for i in bi.deps if not i.noinstall]
    else:
        rpmlist = []
        for dep in bi.deps:
            if dep.sysroot:
                # packages installed in sysroot subdirectory need to get a prefix for init_buildsystem
                rpmlist.append("sysroot: %s %s\n" % (dep.name, dep.fullfilename))
            else:
                rpmlist.append("%s %s\n" % (dep.name, dep.fullfilename))
    for i in imagebins:
        rpmlist.append("%s preinstallimage\n" % i)
    rpmlist += ["%s %s\n" % (i[0], i[1]) for i in rpmlist_prefers]

    if imagefile:
        rpmlist.append('preinstallimage: %s\n' % imagefile)
    if imagesource:
        rpmlist.append('preinstallimagesource: %s\n' % imagesource)

    rpmlist.append('preinstall: ' + ' '.join(bi.preinstall_list) + '\n')
    rpmlist.append('vminstall: ' + ' '.join(bi.vminstall_list) + '\n')
    rpmlist.append('runscripts: ' + ' '.join(bi.runscripts_list) + '\n')
    if build_type != 'kiwi' and build_type != 'docker' and build_type != 'podman' and build_type != 'fissile':
        if bi.noinstall_list:
            rpmlist.append('noinstall: ' + ' '.join(bi.noinstall_list) + '\n')
        if bi.installonly_list:
            rpmlist.append('installonly: ' + ' '.join(bi.installonly_list) + '\n')

    rpmlist_file = NamedTemporaryFile(mode='w+t', prefix='rpmlist.')
    rpmlist_filename = rpmlist_file.name
    rpmlist_file.writelines(rpmlist)
    rpmlist_file.flush()

    subst = {'repo': repo, 'arch': arch, 'project': prj, 'package': pacname}
    vm_options = []
    # XXX check if build-device present
    my_build_device = ''
    if config['build-device']:
        my_build_device = config['build-device'] % subst
    else:
        # obs worker uses /root here but that collides with the
        # /root directory if the build root was used without vm
        # before
        my_build_device = build_root + '/img'

    if vm_type:
        if config['build-swap']:
            my_build_swap = config['build-swap'] % subst
        else:
            my_build_swap = build_root + '/swap'

        vm_options = [f"--vm-type={vm_type}"]
        if vm_telnet:
            vm_options += [f"--vm-telnet={vm_telnet}"]
        if vm_memory:
            vm_options += [f"--memory={vm_memory}"]
        if vm_type != 'lxc' and vm_type != 'nspawn':
            vm_options += [f"--vm-disk={my_build_device}"]
            vm_options += [f"--vm-swap={my_build_swap}"]
            vm_options += [f"--logfile={build_root}/.build.log"]
            if vm_type == 'kvm':
                if config['build-kernel']:
                    vm_options += [f"--vm-kernel={config['build-kernel']}"]
                if config['build-initrd']:
                    vm_options += [f"--vm-initrd={config['build-initrd']}"]

            build_root += '/.mount'
        if vm_disk_size:
            vm_options += [f"--vmdisk-rootsize={vm_disk_size}"]

        if config['build-vmdisk-swapsize']:
            vm_options += [f"--vmdisk-swapsize={config['build-vmdisk-swapsize']}"]
        if config['build-vmdisk-filesystem']:
            vm_options += [f"--vmdisk-filesystem={config['build-vmdisk-filesystem']}"]
        if config['build-vm-user']:
            vm_options += [f"--vm-user={config['build-vm-user']}"]

    if opts.preload:
        print("Preload done for selected repo/arch.")
        sys.exit(0)

    print('Running build')
    cmd = [
        config['build-cmd'],
        f"--root={build_root}",
        f"--rpmlist={rpmlist_filename}",
        f"--dist={bc_filename}",
        f"--arch={bi.buildarch}",
    ]
    cmd += specialcmdopts + vm_options + buildargs
    cmd += [build_descr]

    # determine if we're building under root (user == None) and use su_wrapper accordingly
    if calculate_build_root_user(vm_type) is None:
        cmd = su_wrapper(cmd)

    # change personality, if needed
    if hostarch != bi.buildarch and bi.buildarch in change_personality:
        cmd = [change_personality[bi.buildarch]] + cmd

    # record our settings for later builds
    if not opts.local_package and store.is_package:
        store.last_buildroot = repo, arch, vm_type

    try:
        rc = run_external(cmd[0], *cmd[1:])
        if rc:
            print()
            print(f"Build failed with exit code {rc}")
            print(f"The buildroot was: {build_root}")
            print()
            print("Cleaning the build root may fix the problem or allow you to start debugging from a well-defined state:")
            print("  - add '--clean' option to your 'osc build' command")
            print("  - run 'osc wipe [--vm-type=...]' prior running your 'osc build' command again")
            sys.exit(rc)
    except KeyboardInterrupt as keyboard_interrupt_exception:
        print("keyboard interrupt, killing build ...")
        cmd.append('--kill')
        run_external(cmd[0], *cmd[1:])
        raise keyboard_interrupt_exception

    pacdir = os.path.join(build_root, '.build.packages')
    if os.path.islink(pacdir):
        pacdir = os.readlink(pacdir)
        pacdir = os.path.join(build_root, pacdir)

    if os.path.exists(pacdir):
        (s_built, b_built) = get_built_files(pacdir, bi.buildtype)

        print()
        if s_built:
            print(decode_it(s_built))
        print()
        print(decode_it(b_built))

        if opts.keep_pkgs:
            for i in b_built.splitlines() + s_built.splitlines():
                shutil.copy2(i, os.path.join(opts.keep_pkgs, os.path.basename(decode_it(i))))

    if bi_file:
        bi_file.close()
    if bc_file:
        bc_file.close()
    rpmlist_file.close()

# vim: sw=4 et



from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import


class PackageRevision(XmlModel):
    XML_TAG = "revision"

    rev: int = Field(
        xml_attribute=True,
    )

    vrev: Optional[str] = Field(
        xml_attribute=True,
    )

    srcmd5: str = Field(
    )

    version: str = Field(
    )

    time: int = Field(
    )

    user: str = Field(
    )

    comment: Optional[str] = Field(
    )

    requestid: Optional[int] = Field(
    )

    def get_time_str(self):
        import time
        return time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime(self.time))



import unittest

from osc._private.api import xml_escape


class TestXmlEscape(unittest.TestCase):
    def test_lt(self):
        actual = xml_escape("<")
        expected = "&lt;"
        self.assertEqual(actual, expected)

    def test_gt(self):
        actual = xml_escape(">")
        expected = "&gt;"
        self.assertEqual(actual, expected)

    def test_apos(self):
        actual = xml_escape("'")
        expected = "&apos;"
        self.assertEqual(actual, expected)

    def test_quot(self):
        actual = xml_escape("\"")
        expected = "&quot;"
        self.assertEqual(actual, expected)

    def test_amp(self):
        actual = xml_escape("&")
        expected = "&amp;"
        self.assertEqual(actual, expected)


if __name__ == "__main__":
    unittest.main()



import json
import os
import subprocess
import urllib.parse
from pathlib import Path

from .. import conf as osc_conf
from .. import oscerr


class GitStore:

    @classmethod
    def is_project_dir(cls, path):
        try:
            store = cls(path)
        except oscerr.NoWorkingCopy:
            return False
        return store.is_project

    @classmethod
    def is_package_dir(cls, path):
        try:
            store = cls(path)
        except oscerr.NoWorkingCopy:
            return False
        return store.is_package

    def __init__(self, path, check=True):
        self.path = path
        self.abspath = os.path.abspath(self.path)
        try:
            self.toplevel = self._run_git(["rev-parse", "--show-toplevel"])
            self.toplevel = os.path.abspath(self.toplevel)
        except subprocess.CalledProcessError:
            self.toplevel = None

        # TODO: how to determine if the current git repo contains a project or a package?
        self.is_project = False
        self.is_package = False

        if self.toplevel:
            # NOTE: we have only one store in project-git for all packages
            config_path = os.path.join(self.toplevel, "_config")
            pbuild_path = os.path.join(self.toplevel, "_pbuild")
            if self.toplevel == self.abspath and (os.path.isfile(config_path) or os.path.isfile(pbuild_path)):
                self.is_project = True
                self.is_package = False
            else:
                self.is_project = False
                self.is_package = True

        self._package = None
        self._project = None

        if check and not any([self.is_project, self.is_package]):
            msg = f"Directory '{self.path}' is not a Git SCM working copy"
            raise oscerr.NoWorkingCopy(msg)

        if check and not self.scmurl:
            msg = f"Directory '{self.path}' is a Git SCM repo that lacks the 'origin' remote"
            raise oscerr.NoWorkingCopy(msg)

        # TODO: decide if we need explicit 'git lfs pull' or not
        # self._run_git(["lfs", "pull"])

    def assert_is_project(self):
        if not self.is_project:
            msg = f"Directory '{self.path}' is not a Git SCM working copy of a project"
            raise oscerr.NoWorkingCopy(msg)

    def assert_is_package(self):
        if not self.is_package:
            msg = f"Directory '{self.path}' is not a Git SCM working copy of a package"
            raise oscerr.NoWorkingCopy(msg)

    def _run_git(self, args):
        return subprocess.check_output(["git"] + args, encoding="utf-8", cwd=self.abspath).strip()

    @property
    def apiurl(self):
        from ..obs_scm.store import Store

        # read apiurl from the current directory with .osc metadata
        # NOTE: this never triggers if a store is retrieved from osc.store.get_store(),
        # because obs_scm store takes precedence as .osc is present
        store = Store(self.toplevel, check=False)
        if store.apiurl:
            return store.apiurl

        # read project from parent directory that contains a project with .osc metadata
        store = Store(os.path.join(self.toplevel, ".."), check=False)
        if store.is_project and store.apiurl:
            return store.apiurl

        # HACK: we're using the currently configured apiurl
        return osc_conf.config["apiurl"]

    @property
    def project(self):
        from ..obs_scm.store import Store

        # read project from the current directory with .osc metadata
        # NOTE: this never triggers if a store is retrieved from osc.store.get_store(),
        # because obs_scm store takes precedence as .osc is present
        if self._project is None:
            store = Store(self.toplevel, check=False)
            if store.project:
                self._project = store.project

        # read project from parent directory that contains a project with .osc metadata
        if self._project is None:
            store = Store(os.path.join(self.toplevel, ".."), check=False)
            if store.is_project and store.project:
                self._project = store.project

        # guess project from git branch
        if self._project is None:
            # get project from the branch name
            branch = self._run_git(["branch", "--show-current"])

            # HACK: replace hard-coded mapping with metadata from git or the build service
            # NOTE: you never know which git repo is supposed to be used in which project
            if branch == "factory":
                self._project = "openSUSE:Factory"
            else:
                raise oscerr.NoWorkingCopy(f"Couldn't map git branch '{branch}' to a project")
        return self._project

    @project.setter
    def project(self, value):
        self._project = value

    @property
    def package(self):
        if self._package is None:
            origin = self._run_git(["remote", "get-url", "origin"])
            self._package = Path(urllib.parse.urlsplit(origin).path).stem
        return self._package

    @package.setter
    def package(self, value):
        self._package = value

    def _get_option(self, name):
        try:
            result = self._run_git(["config", "--local", "--get", f"osc.{name}"])
        except subprocess.CalledProcessError:
            result = None
        return result

    def _check_type(self, name, value, expected_type):
        if not isinstance(value, expected_type):
            raise TypeError(f"The option '{name}' should be {expected_type.__name__}, not {type(value).__name__}")

    def _set_option(self, name, value):
        self._run_git(["config", "--local", f"osc.{name}", value])

    def _unset_option(self, name):
        try:
            self._run_git(["config", "--local", "--unset", f"osc.{name}"])
        except subprocess.CalledProcessError:
            pass

    def _get_dict_option(self, name):
        result = self._get_option(name)
        if result is None:
            return None
        result = json.loads(result)
        self._check_type(name, result, dict)
        return result

    def _set_dict_option(self, name, value):
        if value is None:
            self._unset_option(name)
            return
        self._check_type(name, value, dict)
        value = json.dumps(value)
        self._set_option(name, value)

    @property
    def last_buildroot(self):
        self.assert_is_package()
        result = self._get_dict_option("last-buildroot")
        if result is not None:
            result = (result["repo"], result["arch"], result["vm_type"])
        return result

    @last_buildroot.setter
    def last_buildroot(self, value):
        self.assert_is_package()
        if len(value) != 3:
            raise ValueError("A tuple with exactly 3 items is expected: (repo, arch, vm_type)")
        value = {
            "repo": value[0],
            "arch": value[1],
            "vm_type": value[2],
        }
        self._set_dict_option("last-buildroot", value)

    @property
    def scmurl(self):
        try:
            return self._run_git(["remote", "get-url", "origin"])
        except subprocess.CalledProcessError:
            return None



"""
This module implements a lightweight and limited alternative
to pydantic's BaseModel and Field classes.
It works on python 3.6+.

This module IS NOT a supported API, it is meant for osc internal use only.
"""

import copy
import functools
import inspect
import sys
import tempfile
import types
import typing
from typing import Callable
from typing import get_type_hints
from xml.etree import ElementTree as ET

# supported types
from enum import Enum
from typing import Any
from typing import Dict
from typing import List
from typing import NewType
from typing import Optional
from typing import Tuple
from typing import Union

if sys.version_info < (3, 8):

    def get_origin(typ):
        result = getattr(typ, "__origin__", None)
        bases = getattr(result, "__orig_bases__", None)
        if bases:
            result = bases[0]
        return result

else:
    from typing import get_origin

import urllib3.response

from . import xml


__all__ = (
    "BaseModel",
    "XmlModel",
    "Field",
    "NotSet",
    "FromParent",
    "Enum",
    "Dict",
    "List",
    "NewType",
    "Optional",
    "Tuple",
    "Union",
)


class NotSetClass:
    def __repr__(self):
        return "NotSet"

    def __bool__(self):
        return False


NotSet = NotSetClass()


class FromParent(NotSetClass):
    def __init__(self, field_name, *, fallback=NotSet):
        self.field_name = field_name
        self.fallback = fallback

    def __repr__(self):
        return f"FromParent(field_name={self.field_name})"


class Field(property):
    def __init__(
        self,
        default: Any = NotSet,
        description: Optional[str] = None,
        exclude: bool = False,
        get_callback: Optional[Callable] = None,
        **extra,
    ):
        # the default value; it can be a factory function that is lazily evaluated on the first use
        # model sets it to None if it equals to NotSet (for better usability)
        self.default = default

        # a flag indicating, whether the default is a callable with lazy evalution
        self.default_is_lazy = callable(self.default)

        # the name of model's attribute associated with this field instance - set from the model
        self.name = None

        # the type of this field instance - set from the model
        self.type = None

        # the description of the field
        self.description = description

        # docstring - for sphinx and help()
        self.__doc__ = self.description
        if self.__doc__:
            # append information about the default value
            if isinstance(self.default, FromParent):
                self.__doc__ += f"\n\nDefault: inherited from parent config's field ``{self.default.field_name}``"
            elif self.default is not NotSet:
                self.__doc__ += f"\n\nDefault: ``{self.default}``"

        # whether to exclude this field from export
        self.exclude = exclude

        # optional callback to postprocess returned field value
        # it takes (model_instance, value) and returns modified value
        self.get_callback = get_callback

        # extra fields
        self.extra = extra

        # create an instance specific of self.get() so we can annotate it in the model
        self.get_copy = types.FunctionType(
            self.get.__code__,
            self.get.__globals__,
            self.get.__name__,
            self.get.__defaults__,
            self.get.__closure__,
        )
        # turn function into a method by binding it to the instance
        self.get_copy = types.MethodType(self.get_copy, self)

        super().__init__(fget=self.get_copy, fset=self.set, doc=description)

    @property
    def origin_type(self):
        origin_type = get_origin(self.type) or self.type
        if self.is_optional:
            types = [i for i in self.type.__args__ if i != type(None)]
            return get_origin(types[0]) or types[0]
        return origin_type

    @property
    def inner_type(self):
        if self.is_optional:
            types = [i for i in self.type.__args__ if i != type(None)]
            type_ = types[0]
        else:
            type_ = self.type

        if get_origin(type_) != list:
            return None

        if not hasattr(type_, "__args__"):
            return None

        inner_type = [i for i in type_.__args__ if i != type(None)][0]
        return inner_type

    @property
    def is_optional(self):
        origin_type = get_origin(self.type) or self.type
        return origin_type == Union and len(self.type.__args__) == 2 and type(None) in self.type.__args__

    @property
    def is_model(self):
        return inspect.isclass(self.origin_type) and issubclass(self.origin_type, BaseModel)

    @property
    def is_model_list(self):
        return inspect.isclass(self.inner_type) and issubclass(self.inner_type, BaseModel)

    def validate_type(self, value, expected_types=None):
        if not expected_types and self.is_optional and value is None:
            return True

        if expected_types is None:
            expected_types = (self.type,)
        elif not isinstance(expected_types, (list, tuple)):
            expected_types = (expected_types,)

        valid_type = False

        for expected_type in expected_types:
            if valid_type:
                break

            origin_type = get_origin(expected_type) or expected_type

            # unwrap Union
            if origin_type == Union:
                if value is None and type(None) in expected_type.__args__:
                    valid_type = True
                    continue

                valid_type |= self.validate_type(value, expected_types=expected_type.__args__)
                continue

            # unwrap NewType
            if (callable(NewType) or isinstance(origin_type, NewType)) and hasattr(origin_type, "__supertype__"):
                valid_type |= self.validate_type(value, expected_types=(origin_type.__supertype__,))
                continue

            if (
                inspect.isclass(expected_type)
                and issubclass(expected_type, BaseModel)
                and isinstance(value, (expected_type, dict))
            ):
                valid_type = True
                continue

            if (
                inspect.isclass(expected_type)
                and issubclass(expected_type, Enum)
            ):
                # test if the value is part of the enum
                expected_type(value)
                valid_type = True
                continue

            if not isinstance(value, origin_type):
                msg = f"Field '{self.name}' has type '{self.type}'. Cannot assign a value with type '{type(value).__name__}'."
                raise TypeError(msg)

            # the type annotation has no arguments -> no need to check those
            if not getattr(expected_type, "__args__", None):
                valid_type = True
                continue

            if origin_type in (list, tuple):
                valid_type_items = True
                for i in value:
                    valid_type_items &= self.validate_type(i, expected_type.__args__)
                valid_type |= valid_type_items
            elif origin_type in (dict,):
                valid_type_items = True
                for k, v in value.items():
                    valid_type_items &= self.validate_type(k, expected_type.__args__[0])
                    valid_type_items &= self.validate_type(v, expected_type.__args__[1])
                valid_type |= valid_type_items
            else:
                raise TypeError(f"Field '{self.name}' has unsupported type '{self.type}'.")

        return valid_type

    def get(self, obj):
        try:
            result = obj._values[self.name]

            # convert dictionaries into objects
            # we can't do it earlier because list is a standalone object that is not under our control
            if result is not None and self.is_model_list:
                for num, i in enumerate(result):
                    if isinstance(i, dict):
                        klass = self.inner_type
                        result[num] = klass(**i, _parent=obj)

            if self.get_callback is not None:
                result = self.get_callback(obj, result)

            return result
        except KeyError:
            pass

        try:
            result = obj._defaults[self.name]
            if isinstance(result, (dict, list)):
                # make a deepcopy to avoid problems with mutable defaults
                result = copy.deepcopy(result)
                obj._values[self.name] = result
            if self.get_callback is not None:
                result = self.get_callback(obj, result)
            return result
        except KeyError:
            pass

        if isinstance(self.default, FromParent):
            if obj._parent is None:
                if self.default.fallback is not NotSet:
                    return self.default.fallback
                else:
                    raise RuntimeError(f"The field '{self.name}' has default {self.default} but the model has no parent set")
            return getattr(obj._parent, self.default.field_name or self.name)

        if self.default is NotSet:
            raise RuntimeError(f"The field '{self.name}' has no default")

        # make a deepcopy to avoid problems with mutable defaults
        default = copy.deepcopy(self.default)

        # lazy evaluation of a factory function on first use
        if callable(default):
            default = default()

        # if this is a model field, convert dict to a model instance
        if self.is_model and isinstance(default, dict):
            cls = self.origin_type
            new_value = cls()  # pylint: disable=not-callable
            for k, v in default.items():
                setattr(new_value, k, v)
            default = new_value

        obj._defaults[self.name] = default
        return default

    def set(self, obj, value):
        # if this is a model field, convert dict to a model instance
        if self.is_model and isinstance(value, dict):
            # initialize a model instance from a dictionary
            klass = self.origin_type
            value = klass(**value, _parent=obj)  # pylint: disable=not-callable
        elif self.is_model_list and isinstance(value, list):
            new_value = []
            for i in value:
                if isinstance(i, dict):
                    klass = self.inner_type
                    new_value.append(klass(**i, _parent=obj))
                else:
                    i._parent = obj
                    new_value.append(i)
            value = new_value
        elif self.is_model and isinstance(value, str) and hasattr(self.origin_type, "XML_TAG_FIELD"):
            klass = self.origin_type
            key = getattr(self.origin_type, "XML_TAG_FIELD")
            value = klass(**{key: value}, _parent=obj)
        elif self.is_model and value is not None:
            value._parent = obj

        self.validate_type(value)
        obj._values[self.name] = value


class ModelMeta(type):
    def __new__(mcs, name, bases, attrs):
        new_cls = super().__new__(mcs, name, bases, attrs)
        new_cls.__fields__ = {}

        # NOTE: dir() doesn't preserve attribute order
        # we need to iterate through __mro__ classes to workaround that
        for parent_cls in reversed(new_cls.__mro__):
            for field_name in parent_cls.__dict__:
                if field_name in new_cls.__fields__:
                    continue
                field = getattr(new_cls, field_name)
                if not isinstance(field, Field):
                    continue
                new_cls.__fields__[field_name] = field

        # fill model specific details back to the fields
        for field_name, field in new_cls.__fields__.items():
            # property name associated with the field in this model
            field.name = field_name

            # field type associated with the field in this model
            field.type = get_type_hints(new_cls)[field_name]

            # set annotation for the getter so it shows up in sphinx
            field.get_copy.__func__.__annotations__ = {"return": field.type}

            # set 'None' as the default for optional fields
            if field.default is NotSet and field.is_optional:
                field.default = None

        return new_cls


@functools.total_ordering
class BaseModel(metaclass=ModelMeta):
    __fields__: Dict[str, Field]

    def __setattr__(self, name, value):
        if getattr(self, "_allow_new_attributes", True) or hasattr(self.__class__, name) or hasattr(self, name):
            # allow setting properties - test if they exist in the class
            # also allow setting existing attributes that were previously initialized via __dict__
            return super().__setattr__(name, value)
        raise AttributeError(f"Setting attribute '{self.__class__.__name__}.{name}' is not allowed")

    def __init__(self, **kwargs):
        self._allow_new_attributes = True
        self._defaults = {}  # field defaults cached in field.get()
        self._values = {}  # field values explicitly set after initializing the model
        self._parent = kwargs.pop("_parent", None)

        uninitialized_fields = []

        for name, field in self.__fields__.items():
            if name not in kwargs:
                if field.default is NotSet:
                    uninitialized_fields.append(field.name)
                continue
            value = kwargs.pop(name)
            setattr(self, name, value)

        if kwargs:
            unknown_fields_str = ", ".join([f"'{i}'" for i in kwargs])
            raise TypeError(f"The following kwargs of '{self.__class__.__name__}.__init__()' do not match any field: {unknown_fields_str}")

        if uninitialized_fields:
            uninitialized_fields_str = ", ".join([f"'{i}'" for i in uninitialized_fields])
            raise TypeError(
                f"The following fields of '{self.__class__.__name__}' object are not initialized and have no default either: {uninitialized_fields_str}"
            )

        for name, field in self.__fields__.items():
            field.validate_type(getattr(self, name))

        self._snapshot = {}  # copy of ``self.dict()`` so we can determine if the object has changed later on
        self.do_snapshot()

        self._allow_new_attributes = False

    def _get_cmp_data(self):
        result = []
        for name, field in self.__fields__.items():
            if field.exclude:
                continue
            value = getattr(self, name)
            if isinstance(value, dict):
                value = sorted(list(value.items()))
            result.append((name, value))
        return result

    def __eq__(self, other):
        if type(self) != type(other):
            return False
        return self._get_cmp_data() == other._get_cmp_data()

    def __lt__(self, other):
        if type(self) != type(other):
            return False
        return self._get_cmp_data() < other._get_cmp_data()

    def dict(self):
        result = {}
        for name, field in self.__fields__.items():
            if field.exclude:
                continue
            value = getattr(self, name)
            if value is not None and field.is_model:
                result[name] = value.dict()
            elif value is not None and field.is_model_list:
                result[name] = [i.dict() for i in value]
            else:
                result[name] = value

        return result

    def do_snapshot(self):
        """
        Save ``self.dict()`` result as a new starting point for detecting changes in the object data.
        """
        self._snapshot = self.dict()

    def has_changed(self):
        """
        Determine if the object data has changed since its creation or the last snapshot.
        """
        return self.dict() != self._snapshot


class XmlModel(BaseModel):
    XML_TAG = None

    _apiurl: Optional[str] = Field(
        exclude=True,
        default=FromParent("_apiurl", fallback=None),
    )

    def to_xml(self, *, with_comments: bool = False) -> ET.Element:
        xml_tag = None

        # check if there's a special field that sets the tag
        for field_name, field in self.__fields__.items():
            xml_set_tag = field.extra.get("xml_set_tag", False)
            if xml_set_tag:
                value = getattr(self, field_name)
                xml_tag = value
                break

        # use the value from the class
        if xml_tag is None:
            xml_tag = self.XML_TAG

        assert xml_tag is not None
        root = ET.Element(xml_tag)

        if with_comments:
            comment = [
                "",
                "The commented attributes and elements only provide hints on the XML structure.",
                "See OBS documentation such as XML schema files for more details:",
                "https://github.com/openSUSE/open-build-service/tree/master/docs/api/api",
                "",
            ]
            comment_node = ET.Comment(text="\n".join(comment))
            root.append(comment_node)

        for field_name, field in self.__fields__.items():
            if field.exclude:
                continue
            xml_attribute = field.extra.get("xml_attribute", False)
            xml_set_tag = field.extra.get("xml_set_tag", False)
            xml_set_text = field.extra.get("xml_set_text", False)
            xml_name = field.extra.get("xml_name", field_name)
            xml_wrapped = field.extra.get("xml_wrapped", False)
            xml_item_name = field.extra.get("xml_item_name", xml_name)

            if xml_set_tag:
                # a special case when the field determines the top-level tag name
                continue

            if with_comments:
                if xml_attribute:
                    comment = f'{xml_name}=""'
                else:
                    comment = f"<{xml_name}></{xml_name}>"

                comment_node = ET.Comment(text=f" {comment} ")
                root.append(comment_node)

            value = getattr(self, field_name)
            if value is None:
                # skip fields that are not set
                continue

            # if value is wrapped into an external element, create it
            if xml_wrapped:
                wrapper_node = ET.SubElement(root, xml_name)
            else:
                wrapper_node = root

            if xml_set_text:
                wrapper_node.text = str(value)
                continue

            if field.origin_type == list:
                for entry in value:
                    if isinstance(entry, dict):
                        klass = field.inner_type
                        obj = klass(**entry)
                        node = obj.to_xml()
                        wrapper_node.append(node)
                    elif field.inner_type and issubclass(field.inner_type, XmlModel):
                        wrapper_node.append(entry.to_xml())
                    else:
                        node = ET.SubElement(wrapper_node, xml_item_name)
                        if xml_attribute:
                            node.attrib[xml_attribute] = entry
                        else:
                            node.text = entry
            elif issubclass(field.origin_type, XmlModel):
                wrapper_node.append(value.to_xml())
            elif xml_attribute:
                wrapper_node.attrib[xml_name] = str(value)
            else:
                node = ET.SubElement(wrapper_node, xml_name)
                node.text = str(value)
        return root

    @classmethod
    def from_string(cls, string: str, *, apiurl: Optional[str] = None) -> "XmlModel":
        """
        Instantiate model from string.
        """
        root = xml.xml_fromstring(string)
        return cls.from_xml(root, apiurl=apiurl)

    @classmethod
    def from_file(cls, file: Union[str, typing.IO], *, apiurl: Optional[str] = None) -> "XmlModel":
        """
        Instantiate model from file.
        """
        root = xml.xml_parse(file).getroot()
        return cls.from_xml(root, apiurl=apiurl)

    def to_bytes(self, *, with_comments: bool = False) -> bytes:
        """
        Serialize the object as XML and return it as utf-8 encoded bytes.
        """
        root = self.to_xml(with_comments=with_comments)
        xml.xml_indent(root)
        return ET.tostring(root, encoding="utf-8")

    def to_string(self, *, with_comments: bool = False) -> str:
        """
        Serialize the object as XML and return it as a string.
        """
        return self.to_bytes(with_comments=with_comments).decode("utf-8")

    def to_file(self, file: Union[str, typing.IO], *, with_comments: bool = False) -> None:
        """
        Serialize the object as XML and save it to an utf-8 encoded file.
        """
        root = self.to_xml(with_comments=with_comments)
        xml.xml_indent(root)
        return ET.ElementTree(root).write(file, encoding="utf-8")

    @staticmethod
    def value_from_string(field, value):
        """
        Convert field value from string to the actual type of the field.
        """
        if field.origin_type is bool:
            if value.lower() in ["1", "yes", "true", "on"]:
                value = True
                return value
            if value.lower() in ["0", "no", "false", "off"]:
                value = False
                return value

        if field.origin_type is int:
            if not value or not value.strip():
                return None
            value = int(value)
            return value

        return value

    @classmethod
    def _remove_processed_node(cls, parent, node):
        """
        Remove a node that has been fully processed and is now empty.
        """
        if len(node) != 0:
            raise RuntimeError(f"Node {node} contains unprocessed child elements {list(node)}")
        if node.attrib:
            raise RuntimeError(f"Node {node} contains unprocessed attributes {node.attrib}")
        if node.text is not None and node.text.strip():
            raise RuntimeError(f"Node {node} contains unprocessed text {node.text}")
        if parent is not None:
            parent.remove(node)

    @classmethod
    def from_xml(cls, root: ET.Element, *, apiurl: Optional[str] = None):
        """
        Instantiate model from a XML root.
        """

        # We need to make sure we parse all data
        # and that's why we remove processed elements and attributes and check that nothing remains.
        # Otherwise we'd be sending partial XML back and that would lead to data loss.
        #
        # Let's make a copy of the xml tree because we'll destroy it during the process.
        orig_root = root
        root = copy.deepcopy(root)

        kwargs = {}
        for field_name, field in cls.__fields__.items():
            xml_attribute = field.extra.get("xml_attribute", False)
            xml_set_tag = field.extra.get("xml_set_tag", False)
            xml_set_text = field.extra.get("xml_set_text", False)
            xml_name = field.extra.get("xml_name", field_name)
            xml_wrapped = field.extra.get("xml_wrapped", False)
            xml_item_name = field.extra.get("xml_item_name", xml_name)
            value: Any
            node: Optional[ET.Element]

            if xml_set_tag:
                # field contains name of the ``root`` tag
                if xml_wrapped:
                    # the last node wins (overrides the previous nodes)
                    for node in root[:]:
                        value = node.tag
                        cls._remove_processed_node(root, node)
                else:
                    value = root.tag

                kwargs[field_name] = value
                continue

            if xml_set_text:
                # field contains the value (text) of the element
                if xml_wrapped:
                    # the last node wins (overrides the previous nodes)
                    for node in root[:]:
                        value = node.text
                        node.text = None
                        cls._remove_processed_node(root, node)
                else:
                    value = root.text
                    root.text = None

                value = value.strip()
                kwargs[field_name] = value
                continue

            if xml_attribute:
                # field is an attribute that contains a scalar
                if xml_name not in root.attrib:
                    continue
                value = cls.value_from_string(field, root.attrib.pop(xml_name))
                kwargs[field_name] = value
                continue

            if field.origin_type is list:
                if xml_wrapped:
                    wrapper_node = root.find(xml_name)
                    # we'll consider all nodes inside the wrapper node
                    nodes = wrapper_node[:] if wrapper_node is not None else None
                else:
                    wrapper_node = None
                    # we'll consider only nodes with matching name
                    nodes = root.findall(xml_item_name)

                if not nodes:
                    if wrapper_node is not None:
                        cls._remove_processed_node(root, wrapper_node)
                    continue

                values = []
                for node in nodes:
                    if field.is_model_list:
                        klass = field.inner_type
                        entry = klass.from_xml(node, apiurl=apiurl)

                        # clear node as it was checked in from_xml() already
                        node.text = None
                        node.attrib = {}
                        node[:] = []
                    else:
                        entry = cls.value_from_string(field, node.text)
                        node.text = None

                    values.append(entry)

                    if xml_wrapped:
                        cls._remove_processed_node(wrapper_node, node)
                    else:
                        cls._remove_processed_node(root, node)

                if xml_wrapped:
                    cls._remove_processed_node(root, wrapper_node)

                kwargs[field_name] = values
                continue

            if field.is_model:
                # field contains an instance of XmlModel
                assert xml_name is not None
                node = root.find(xml_name)
                if node is None:
                    continue
                klass = field.origin_type
                kwargs[field_name] = klass.from_xml(node, apiurl=apiurl)

                # clear node as it was checked in from_xml() already
                node.text = None
                node.attrib = {}
                node[:] = []

                cls._remove_processed_node(root, node)
                continue

            # field contains a scalar
            node = root.find(xml_name)
            if node is None:
                continue
            value = cls.value_from_string(field, node.text)
            node.text = None
            cls._remove_processed_node(root, node)
            if value is None:
                if field.is_optional:
                    continue
                value = ""
            kwargs[field_name] = value

        cls._remove_processed_node(None, root)

        obj = cls(**kwargs, _apiurl=apiurl)
        obj.__dict__["_root"] = orig_root
        return obj

    @classmethod
    def xml_request(
        cls,
        method: str,
        apiurl: str,
        path: List[str],
        query: Optional[dict] = None,
        headers: Optional[str] = None,
        data: Optional[str] = None,
    ) -> urllib3.response.HTTPResponse:
        from ..connection import http_request
        from ..core import makeurl
        url = makeurl(apiurl, path, query)
        # TODO: catch HTTPError and return the wrapped response as XmlModel instance
        return http_request(method, url, headers=headers, data=data)

    def do_update(self, other: "XmlModel") -> None:
        """
        Update values of the fields in the current model instance from another.
        """
        self._values = copy.deepcopy(other._values)

    def do_edit(self) -> Tuple[str, str, "XmlModel"]:
        """
        Serialize model as XML and open it in an editor for editing.
        Return a tuple with:
          * a string with original data
          * a string with edited data
          * an instance of the class with edited data loaded

        IMPORTANT: This method is always interactive.
        """
        from ..core import run_editor
        from ..output import get_user_input

        def write_file(f, data):
            f.seek(0)
            f.write(data)
            f.truncate()
            f.flush()

        with tempfile.NamedTemporaryFile(mode="w+", encoding="utf-8", prefix="obs_xml_", suffix=".xml") as f:
            original_data = self.to_string()
            original_data_with_comments = self.to_string(with_comments=True)
            write_file(f, original_data_with_comments)

            while True:
                run_editor(f.name)
                try:
                    edited_obj = self.__class__.from_file(f.name, apiurl=self._apiurl)
                    f.seek(0)
                    edited_data = f.read()
                    break
                except Exception as e:
                    reply = get_user_input(
                        f"""
                        The edited data is not valid.
                        {e}
                        """,
                        answers={"a": "abort", "e": "edit", "u": "undo changes and edit"},
                    )
                    if reply == "a":
                        from .. import oscerr
                        raise oscerr.UserAbort()
                    elif reply == "e":
                        continue
                    elif reply == "u":
                        write_file(f, original_data_with_comments)
                        continue

        # strip comments, we don't need to increase traffic to the server
        edited_data = edited_obj.to_string()

        return original_data, edited_data, edited_obj



import fnmatch
import os
from pathlib import Path
from typing import Optional

from .. import conf
from .. import oscerr
from ..util.xml import ET
from ..util.xml import xml_parse
from .store import Store
from .store import delete_storedir
from .store import store
from .store import store_read_package
from .store import store_read_project
from .store import store_write_initial_packages
from .store import store_write_project
from .store import store_write_string
from .store import is_package_dir


class Project:
    """
    Represent a checked out project directory, holding packages.

    :Attributes:
        ``dir``
            The directory path containing the project.

        ``name``
            The name of the project.

        ``apiurl``
            The endpoint URL of the API server.

        ``pacs_available``
            List of names of packages available server-side.
            This is only populated if ``getPackageList`` is set
            to ``True`` in the constructor.

        ``pacs_have``
            List of names of packages which exist server-side
            and exist in the local project working copy (if
            'do_package_tracking' is disabled).
            If 'do_package_tracking' is enabled it represents the
            list names of packages which are tracked in the project
            working copy (that is it might contain packages which
            exist on the server as well as packages which do not
            exist on the server (for instance if the local package
            was added or if the package was removed on the server-side)).

        ``pacs_excluded``
            List of names of packages in the local project directory
            which are excluded by the `exclude_glob` configuration
            variable.  Only set if `do_package_tracking` is enabled.

        ``pacs_unvers``
            List of names of packages in the local project directory
            which are not tracked. Only set if `do_package_tracking`
            is enabled.

        ``pacs_broken``
            List of names of packages which are tracked but do not
            exist in the local project working copy. Only set if
            `do_package_tracking` is enabled.

        ``pacs_missing``
            List of names of packages which exist server-side but
            are not expected to exist in the local project directory.
    """

    REQ_STOREFILES = ('_project', '_apiurl')

    def __init__(self, dir, getPackageList=True, progress_obj=None, wc_check=True):
        """
        Constructor.

        :Parameters:
            `dir` : str
                The directory path containing the checked out project.

            `getPackageList` : bool
                Set to `False` if you want to skip retrieval from the
                server of the list of packages in the project .

            `wc_check` : bool
        """
        from ..core import meta_get_packagelist

        self.dir = Path(dir)
        self.absdir = os.path.abspath(dir)
        self.store = Store(dir, check=wc_check)
        self.progress_obj = progress_obj

        self.name = store_read_project(self.dir)
        self.scm_url = self.store.scmurl
        self.apiurl = self.store.apiurl

        dirty_files = []
        if wc_check:
            dirty_files = self.wc_check()
        if dirty_files:
            msg = 'Your working copy \'%s\' is in an inconsistent state.\n' \
                'Please run \'osc repairwc %s\' and check the state\n' \
                'of the working copy afterwards (via \'osc status %s\')' % (self.dir, self.dir, self.dir)
            raise oscerr.WorkingCopyInconsistent(self.name, None, dirty_files, msg)

        if getPackageList:
            self.pacs_available = meta_get_packagelist(self.apiurl, self.name)
        else:
            self.pacs_available = []

        if conf.config['do_package_tracking']:
            self.pac_root = self.read_packages().getroot()
            self.pacs_have = [pac.get('name') for pac in self.pac_root.findall('package')]
            self.pacs_excluded = [i for i in os.listdir(self.dir)
                                  for j in conf.config['exclude_glob']
                                  if fnmatch.fnmatch(i, j)]
            self.pacs_unvers = [i for i in os.listdir(self.dir) if i not in self.pacs_have and i not in self.pacs_excluded]
            # store all broken packages (e.g. packages which where removed by a non-osc cmd)
            # in the self.pacs_broken list
            self.pacs_broken = []
            for p in self.pacs_have:
                if not os.path.isdir(os.path.join(self.absdir, p)):
                    # all states will be replaced with the '!'-state
                    # (except it is already marked as deleted ('D'-state))
                    self.pacs_broken.append(p)
        else:
            self.pacs_have = [i for i in os.listdir(self.dir) if i in self.pacs_available]

        self.pacs_missing = [i for i in self.pacs_available if i not in self.pacs_have]

    def wc_check(self):
        global store
        dirty_files = []
        req_storefiles = Project.REQ_STOREFILES
        if conf.config['do_package_tracking'] and self.scm_url is None:
            req_storefiles += ('_packages',)
        for fname in req_storefiles:
            if not os.path.exists(os.path.join(self.absdir, store, fname)):
                dirty_files.append(fname)
        return dirty_files

    def wc_repair(self, apiurl: Optional[str] = None) -> bool:
        repaired: bool = False

        store = Store(self.dir, check=False)
        store.assert_is_project()

        # there was a time when osc did not write _osclib_version file; let's assume these checkouts have version 1.0
        if not store.exists("_osclib_version"):
            store.write_string("_osclib_version", "1.0")
            repaired = True

        if not store.exists("_apiurl") or apiurl:
            if apiurl is None:
                msg = 'cannot repair wc: the \'_apiurl\' file is missing but ' \
                    'no \'apiurl\' was passed to wc_repair'
                # hmm should we raise oscerr.WrongArgs?
                raise oscerr.WorkingCopyInconsistent(self.name, None, [], msg)
            # sanity check
            conf.parse_apisrv_url(None, apiurl)
            store.apiurl = apiurl
            self.apiurl = apiurl
            repaired = True

        return repaired

    def checkout_missing_pacs(self, sinfos, expand_link=False, unexpand_link=False):
        from ..core import checkout_package
        from ..core import getTransActPath

        for pac in self.pacs_missing:
            if conf.config['do_package_tracking'] and pac in self.pacs_unvers:
                # pac is not under version control but a local file/dir exists
                msg = f'can\'t add package \'{pac}\': Object already exists'
                raise oscerr.PackageExists(self.name, pac, msg)

            if not (expand_link or unexpand_link):
                sinfo = sinfos.get(pac)
                if sinfo is None:
                    # should never happen...
                    continue
                linked = sinfo.find('linked')
                if linked is not None and linked.get('project') == self.name:
                    # hmm what about a linkerror (sinfo.get('lsrcmd5') is None)?
                    # Should we skip the package as well or should we it out?
                    # let's skip it for now
                    print(f"Skipping {pac} (link to package {linked.get('package')})")
                    continue

            print(f'checking out new package {pac}')
            checkout_package(self.apiurl, self.name, pac,
                             pathname=getTransActPath(os.path.join(self.dir, pac)),
                             prj_obj=self, prj_dir=self.dir,
                             expand_link=expand_link or not unexpand_link, progress_obj=self.progress_obj)

    def status(self, pac: str):
        exists = os.path.exists(os.path.join(self.absdir, pac))
        st = self.get_state(pac)
        if st is None and exists:
            return '?'
        elif st is None:
            raise oscerr.OscIOError(None, f'osc: \'{pac}\' is not under version control')
        elif st in ('A', ' ') and not exists:
            return '!'
        elif st == 'D' and not exists:
            return 'D'
        else:
            return st

    def get_status(self, *exclude_states):
        res = []
        for pac in self.pacs_have:
            st = self.status(pac)
            if st not in exclude_states:
                res.append((st, pac))
        if '?' not in exclude_states:
            res.extend([('?', pac) for pac in self.pacs_unvers])
        return res

    def get_pacobj(self, pac, *pac_args, **pac_kwargs):
        from ..core import Package

        try:
            st = self.status(pac)
            if st in ('?', '!') or st == 'D' and not os.path.exists(os.path.join(self.dir, pac)):
                return None
            return Package(os.path.join(self.dir, pac), *pac_args, **pac_kwargs)
        except oscerr.OscIOError:
            return None

    def set_state(self, pac, state):
        node = self.get_package_node(pac)
        if node is None:
            self.new_package_entry(pac, state)
        else:
            node.set('state', state)

    def get_package_node(self, pac: str):
        for node in self.pac_root.findall('package'):
            if pac == node.get('name'):
                return node
        return None

    def del_package_node(self, pac):
        for node in self.pac_root.findall('package'):
            if pac == node.get('name'):
                self.pac_root.remove(node)

    def get_state(self, pac: str):
        node = self.get_package_node(pac)
        if node is not None:
            return node.get('state')
        else:
            return None

    def info(self):
        from ..core import project_info_templ
        from ..core import makeurl

        source_url = makeurl(self.apiurl, ['source', self.name])
        r = project_info_templ % (self.name, self.absdir, self.apiurl, source_url)
        return r

    def new_package_entry(self, name, state):
        ET.SubElement(self.pac_root, 'package', name=name, state=state)

    def read_packages(self):
        """
        Returns an ``xml.etree.ElementTree`` object representing the
        parsed contents of the project's ``.osc/_packages`` XML file.
        """
        from ..core import Package
        from ..core import meta_get_packagelist

        global store

        packages_file = os.path.join(self.absdir, store, '_packages')
        if os.path.isfile(packages_file) and os.path.getsize(packages_file):
            try:
                result = xml_parse(packages_file)
            except:
                msg = f'Cannot read package file \'{packages_file}\'. '
                msg += 'You can try to remove it and then run osc repairwc.'
                raise oscerr.OscIOError(None, msg)
            return result
        else:
            # scan project for existing packages and migrate them
            cur_pacs = []
            for data in os.listdir(self.dir):
                pac_dir = os.path.join(self.absdir, data)
                # we cannot use self.pacs_available because we cannot guarantee that the package list
                # was fetched from the server
                if data in meta_get_packagelist(self.apiurl, self.name) and is_package_dir(pac_dir) \
                   and Package(pac_dir).name == data:
                    cur_pacs.append(ET.Element('package', name=data, state=' '))
            store_write_initial_packages(self.absdir, self.name, cur_pacs)
            return xml_parse(os.path.join(self.absdir, store, '_packages'))

    def write_packages(self):
        from ..core import ET_ENCODING
        from ..core import xmlindent

        xmlindent(self.pac_root)
        store_write_string(self.absdir, '_packages', ET.tostring(self.pac_root, encoding=ET_ENCODING))

    def addPackage(self, pac):
        for i in conf.config['exclude_glob']:
            if fnmatch.fnmatch(pac, i):
                msg = f'invalid package name: \'{pac}\' (see \'exclude_glob\' config option)'
                raise oscerr.OscIOError(None, msg)
        state = self.get_state(pac)
        if state is None or state == 'D':
            self.new_package_entry(pac, 'A')
            self.write_packages()
            # sometimes the new pac doesn't exist in the list because
            # it would take too much time to update all data structs regularly
            if pac in self.pacs_unvers:
                self.pacs_unvers.remove(pac)
        else:
            raise oscerr.PackageExists(self.name, pac, f'package \'{pac}\' is already under version control')

    def delPackage(self, pac, force=False):
        from ..core import delete_dir
        from ..core import getTransActPath
        from ..core import statfrmt

        state = self.get_state(pac.name)
        can_delete = True
        if state == ' ' or state == 'D':
            del_files = []
            for filename in pac.filenamelist + pac.filenamelist_unvers:
                filestate = pac.status(filename)
                if filestate == 'M' or filestate == 'C' or \
                   filestate == 'A' or filestate == '?':
                    can_delete = False
                else:
                    del_files.append(filename)
            if can_delete or force:
                for filename in del_files:
                    pac.delete_localfile(filename)
                    if pac.status(filename) != '?':
                        # this is not really necessary
                        pac.put_on_deletelist(filename)
                        print(statfrmt('D', getTransActPath(os.path.join(pac.dir, filename))))
                print(statfrmt('D', getTransActPath(os.path.join(pac.dir, os.pardir, pac.name))))
                pac.write_deletelist()
                self.set_state(pac.name, 'D')
                self.write_packages()
            else:
                print(f'package \'{pac.name}\' has local modifications (see osc st for details)')
        elif state == 'A':
            if force:
                delete_dir(pac.absdir)
                self.del_package_node(pac.name)
                self.write_packages()
                print(statfrmt('D', pac.name))
            else:
                print(f'package \'{pac.name}\' has local modifications (see osc st for details)')
        elif state is None:
            print('package is not under version control')
        else:
            print('unsupported state')

    def update(self, pacs=(), expand_link=False, unexpand_link=False, service_files=False):
        from ..core import Package
        from ..core import checkout_package
        from ..core import get_project_sourceinfo
        from ..core import getTransActPath
        from ..core import show_upstream_xsrcmd5

        if pacs:
            for pac in pacs:
                Package(os.path.join(self.dir, pac), progress_obj=self.progress_obj).update()
        else:
            # we need to make sure that the _packages file will be written (even if an exception
            # occurs)
            try:
                # update complete project
                # packages which no longer exists upstream
                upstream_del = [pac for pac in self.pacs_have if pac not in self.pacs_available and self.get_state(pac) != 'A']
                sinfo_pacs = [pac for pac in self.pacs_have if self.get_state(pac) in (' ', 'D') and pac not in self.pacs_broken]
                sinfo_pacs.extend(self.pacs_missing)
                sinfos = get_project_sourceinfo(self.apiurl, self.name, True, *sinfo_pacs)

                for pac in upstream_del:
                    if self.status(pac) != '!':
                        p = Package(os.path.join(self.dir, pac))
                        self.delPackage(p, force=True)
                        delete_storedir(p.storedir)
                        try:
                            os.rmdir(pac)
                        except:
                            pass
                    self.pac_root.remove(self.get_package_node(pac))
                    self.pacs_have.remove(pac)

                for pac in self.pacs_have:
                    state = self.get_state(pac)
                    if pac in self.pacs_broken:
                        if self.get_state(pac) != 'A':
                            checkout_package(self.apiurl, self.name, pac,
                                             pathname=getTransActPath(os.path.join(self.dir, pac)), prj_obj=self,
                                             prj_dir=self.dir, expand_link=not unexpand_link, progress_obj=self.progress_obj)
                    elif state == ' ':
                        # do a simple update
                        p = Package(os.path.join(self.dir, pac), progress_obj=self.progress_obj)
                        rev = None
                        needs_update = True
                        if p.scm_url is not None:
                            # git managed.
                            print("Skipping git managed package ", pac)
                            continue
                        elif expand_link and p.islink() and not p.isexpanded():
                            if p.haslinkerror():
                                try:
                                    rev = show_upstream_xsrcmd5(p.apiurl, p.prjname, p.name, revision=p.rev)
                                except:
                                    rev = show_upstream_xsrcmd5(p.apiurl, p.prjname, p.name, revision=p.rev, linkrev="base")
                                    p.mark_frozen()
                            else:
                                rev = p.linkinfo.xsrcmd5
                            print('Expanding to rev', rev)
                        elif unexpand_link and p.islink() and p.isexpanded():
                            rev = p.linkinfo.lsrcmd5
                            print('Unexpanding to rev', rev)
                        elif p.islink() and p.isexpanded():
                            needs_update = p.update_needed(sinfos[p.name])
                            if needs_update:
                                rev = p.latest_rev()
                        elif p.hasserviceinfo() and p.serviceinfo.isexpanded() and not service_files:
                            # FIXME: currently, do_update does not propagate the --server-side-source-service-files
                            # option to this method. Consequence: an expanded service is always unexpanded during
                            # an update (TODO: discuss if this is a reasonable behavior (at least this the default
                            # behavior for a while))
                            needs_update = True
                        else:
                            needs_update = p.update_needed(sinfos[p.name])
                        print(f'Updating {p.name}')
                        if needs_update:
                            p.update(rev, service_files)
                        else:
                            print(f'At revision {p.rev}.')
                        if unexpand_link:
                            p.unmark_frozen()
                    elif state == 'D':
                        # pac exists (the non-existent pac case was handled in the first if block)
                        p = Package(os.path.join(self.dir, pac), progress_obj=self.progress_obj)
                        if p.update_needed(sinfos[p.name]):
                            p.update()
                    elif state == 'A' and pac in self.pacs_available:
                        # file/dir called pac already exists and is under version control
                        msg = f'can\'t add package \'{pac}\': Object already exists'
                        raise oscerr.PackageExists(self.name, pac, msg)
                    elif state == 'A':
                        # do nothing
                        pass
                    else:
                        print(f'unexpected state.. package \'{pac}\'')

                self.checkout_missing_pacs(sinfos, expand_link, unexpand_link)
            finally:
                self.write_packages()

    def commit(self, pacs=(), msg='', files=None, verbose=False, skip_local_service_run=False, can_branch=False, force=False):
        from ..core import Package
        from ..core import os_path_samefile

        files = files or {}
        if pacs:
            try:
                for pac in pacs:
                    todo = []
                    if pac in files:
                        todo = files[pac]
                    state = self.get_state(pac)
                    if state == 'A':
                        self.commitNewPackage(pac, msg, todo, verbose=verbose, skip_local_service_run=skip_local_service_run)
                    elif state == 'D':
                        self.commitDelPackage(pac, force=force)
                    elif state == ' ':
                        # display the correct dir when sending the changes
                        if os_path_samefile(os.path.join(self.dir, pac), os.getcwd()):
                            p = Package('.')
                        else:
                            p = Package(os.path.join(self.dir, pac))
                        p.todo = todo
                        p.commit(msg, verbose=verbose, skip_local_service_run=skip_local_service_run, can_branch=can_branch, force=force)
                    elif pac in self.pacs_unvers and not is_package_dir(os.path.join(self.dir, pac)):
                        print(f'osc: \'{pac}\' is not under version control')
                    elif pac in self.pacs_broken or not os.path.exists(os.path.join(self.dir, pac)):
                        print(f'osc: \'{pac}\' package not found')
                    elif state is None:
                        self.commitExtPackage(pac, msg, todo, verbose=verbose, skip_local_service_run=skip_local_service_run)
            finally:
                self.write_packages()
        else:
            # if we have packages marked as '!' we cannot commit
            for pac in self.pacs_broken:
                if self.get_state(pac) != 'D':
                    msg = f'commit failed: package \'{pac}\' is missing'
                    raise oscerr.PackageMissing(self.name, pac, msg)
            try:
                for pac in self.pacs_have:
                    state = self.get_state(pac)
                    if state == ' ':
                        # do a simple commit
                        Package(os.path.join(self.dir, pac)).commit(msg, verbose=verbose, skip_local_service_run=skip_local_service_run)
                    elif state == 'D':
                        self.commitDelPackage(pac, force=force)
                    elif state == 'A':
                        self.commitNewPackage(pac, msg, verbose=verbose, skip_local_service_run=skip_local_service_run)
            finally:
                self.write_packages()

    def commitNewPackage(self, pac, msg='', files=None, verbose=False, skip_local_service_run=False):
        """creates and commits a new package if it does not exist on the server"""
        from ..core import Package
        from ..core import edit_meta
        from ..core import os_path_samefile
        from ..core import statfrmt

        files = files or []
        if pac in self.pacs_available:
            print(f'package \'{pac}\' already exists')
        else:
            user = conf.get_apiurl_usr(self.apiurl)
            edit_meta(metatype='pkg',
                      path_args=(self.name, pac),
                      template_args=({
                          'name': pac,
                          'user': user}),
                      apiurl=self.apiurl)
            # display the correct dir when sending the changes
            olddir = os.getcwd()
            if os_path_samefile(os.path.join(self.dir, pac), os.curdir):
                os.chdir(os.pardir)
                p = Package(pac)
            else:
                p = Package(os.path.join(self.dir, pac))
            p.todo = files
            print(statfrmt('Sending', os.path.normpath(p.dir)))
            p.commit(msg=msg, verbose=verbose, skip_local_service_run=skip_local_service_run)
            self.set_state(pac, ' ')
            os.chdir(olddir)

    def commitDelPackage(self, pac, force=False):
        """deletes a package on the server and in the working copy"""

        from ..core import Package
        from ..core import delete_package
        from ..core import getTransActPath
        from ..core import os_path_samefile
        from ..core import statfrmt

        try:
            # display the correct dir when sending the changes
            if os_path_samefile(os.path.join(self.dir, pac), os.curdir):
                pac_dir = pac
            else:
                pac_dir = os.path.join(self.dir, pac)
            p = Package(os.path.join(self.dir, pac))
            # print statfrmt('Deleting', os.path.normpath(os.path.join(p.dir, os.pardir, pac)))
            delete_storedir(p.storedir)
            try:
                os.rmdir(p.dir)
            except:
                pass
        except OSError:
            pac_dir = os.path.join(self.dir, pac)
        except (oscerr.NoWorkingCopy, oscerr.WorkingCopyOutdated, oscerr.PackageError):
            pass
        # print statfrmt('Deleting', getTransActPath(os.path.join(self.dir, pac)))
        print(statfrmt('Deleting', getTransActPath(pac_dir)))
        delete_package(self.apiurl, self.name, pac, force=force)
        self.del_package_node(pac)

    def commitExtPackage(self, pac, msg, files=None, verbose=False, skip_local_service_run=False):
        """commits a package from an external project"""

        from ..core import Package
        from ..core import edit_meta
        from ..core import meta_exists
        from ..core import os_path_samefile

        files = files or []
        if os_path_samefile(os.path.join(self.dir, pac), os.getcwd()):
            pac_path = '.'
        else:
            pac_path = os.path.join(self.dir, pac)

        store = Store(pac_path)
        project = store_read_project(pac_path)
        package = store_read_package(pac_path)
        apiurl = store.apiurl
        if not meta_exists(metatype='pkg',
                           path_args=(project, package),
                           template_args=None, create_new=False, apiurl=apiurl):
            user = conf.get_apiurl_usr(self.apiurl)
            edit_meta(metatype='pkg',
                      path_args=(project, package),
                      template_args=({'name': pac, 'user': user}), apiurl=apiurl)
        p = Package(pac_path)
        p.todo = files
        p.commit(msg=msg, verbose=verbose, skip_local_service_run=skip_local_service_run)

    def __str__(self):
        r = []
        r.append('*****************************************************')
        r.append(f'Project {self.name} (dir={self.dir}, absdir={self.absdir})')
        r.append(f"have pacs:\n{', '.join(self.pacs_have)}")
        r.append(f"missing pacs:\n{', '.join(self.pacs_missing)}")
        r.append('*****************************************************')
        return '\n'.join(r)

    @staticmethod
    def init_project(
        apiurl: str,
        dir: Path,
        project,
        package_tracking=True,
        getPackageList=True,
        progress_obj=None,
        wc_check=True,
        scm_url=None,
    ):
        global store

        if not os.path.exists(dir):
            # use makedirs (checkout_no_colon config option might be enabled)
            os.makedirs(dir)
        elif not os.path.isdir(dir):
            raise oscerr.OscIOError(None, f'error: \'{dir}\' is no directory')
        if os.path.exists(os.path.join(dir, store)):
            raise oscerr.OscIOError(None, f'error: \'{dir}\' is already an initialized osc working copy')
        else:
            os.mkdir(os.path.join(dir, store))

        s = Store(dir, check=False)
        s.write_string("_osclib_version", Store.STORE_VERSION)
        s.apiurl = apiurl
        s.project = project
        if scm_url:
            s.scmurl = scm_url
            package_tracking = None
        if package_tracking:
            store_write_initial_packages(dir, project, [])
        return Project(dir, getPackageList, progress_obj, wc_check)
